""" TempBucketOutputFiles model
"""
from django_mongoengine import Document, fields
from mongoengine import errors as mongoengine_errors

from core_custom_queries_app.components.temp_bucket_output_file.models import TempBucketOutputFiles
from core_main_app.commons import exceptions


class TempOutputFile(Document):
    """
    Output files
    """
    data_type = fields.StringField()  # Type of the file (cxv, json or xml)
    # FIXME: use GridFS?
    file = fields.ListField(fields.ReferenceField(TempBucketOutputFiles, blank=True), blank=True)  # List of file parts

    @staticmethod
    def get_by_id(temp_output_file_id):
        """Return a TempOutputFile given its id.

        Args:
            temp_output_file_id:

        Returns:

        """
        try:
            return TempOutputFile.objects.get(pk=str(temp_output_file_id))
        except mongoengine_errors.DoesNotExist as e:
            raise exceptions.DoesNotExist(e.message)
        except Exception as ex:
            raise exceptions.ModelError(ex.message)

    @staticmethod
    def get_all():
        """Return all TempOutputFile.

        Returns:

        """
        return TempOutputFile.objects().all()

    def delete_database(self):
        """
        Delete function

        Delete all part of the file.
        """
        for part_file in self.file:
            part_file.delete_database()
        self.delete()

    def save_file(self, str_file, type_file):
        """
        Save the current file.

        :param str_file: File content
        :param type_file: Type of file
        :return: id of the saved file
        """

        # Cut the file
        chunks_file = [str_file[x:x + 11000000] for x in xrange(0, len(str_file), 11000000)]
        self.data_type = type_file
        list_chunks = list()
        # Save each part of the file
        for chunk_file in chunks_file:
            temp_bucket_output_file = TempBucketOutputFiles()
            temp_bucket_output_file.piece_file = chunk_file
            temp_bucket_output_file.save()
            list_chunks.append(temp_bucket_output_file)
        self.file = list_chunks
        self.save()
        return str(self.id)

    @staticmethod
    def load_file(id_output_file):
        """
        Load the output file.

        :param id_output_file: File id to load.
        :return: File content
        """
        temp_output_file = TempOutputFile.get_by_id(id_output_file)

        list_output_file = []
        for chunk in temp_output_file.file:  # Load each part of the file
            list_output_file.append(chunk.piece_file)
        return "".join(list_output_file)  # Create the file



""" TempChoiceListFile api.
"""
from core_custom_queries_app.components.temp_choice_list_file.models import TempChoiceListFile


def get_all():
    """Return all TempChoiceListFile.

    Returns:

    """
    return TempChoiceListFile.get_all()



""" Log File api
"""
from core_custom_queries_app.components.log_file.models import LogFile


def upsert(log_file):
    """

    Args:
        log_file:

    Returns:

    """
    return log_file.save()


def get_by_id(log_file_id):
    """Return a LogFile given its id.

    Args:
        log_file_id:

    Returns:

    """
    return LogFile.get_by_id(log_file_id)


def get_all():
    """Return all LogFiles.

    Returns:

    """
    return LogFile.get_all()




""" Custom Queries models
"""
# FIXME: these models cannot be split into components for now because of circular dependencies
from datetime import datetime
from string import maketrans

from bson import ObjectId
from django.contrib.auth.models import User
from django.db import models
from django_mongoengine import fields, Document
from mongoengine import errors as mongoengine_errors
from redis import Redis, ConnectionError

from core_custom_queries_app.components.dyn_query.models import DynQuery
from core_custom_queries_app.components.log_file import api as log_file_api
from core_custom_queries_app.components.log_file.models import LogFile
from core_custom_queries_app.components.temp_bucket_id_files.models import TempBucketIdFiles
from core_custom_queries_app.components.temp_choice_list_file.models import TempChoiceListFile
from core_custom_queries_app.components.temp_output_file.models import TempOutputFile
from core_custom_queries_app.components.temp_user_step.models import TempUserStep
from core_custom_queries_app.exceptions import EmptyChoicesFromQuery
from core_custom_queries_app.permissions import rights
from core_custom_queries_app.utils import possible_projection, get_common_key_and_specific_header, get_title_data_leaf, \
    print_bloc, get_header_parents, get_general_key_output_dictionary, flat_list, send_mail_query_end, explore_star
from core_main_app.commons import exceptions
from core_main_app.commons.exceptions import DoesNotExist, ModelError
from core_main_app.components.template.models import Template
from core_main_app.permissions.utils import get_formatted_name
from core_main_app.system import api as system_api
from qdr.settings import REDIS_URL


class CustomQueries(models.Model):
    class Meta:
        verbose_name = 'core_custom_queries_app'
        default_permissions = ()
        permissions = (
            (rights.custom_queries_access, get_formatted_name(rights.custom_queries_access)),
        )


class HistoryQuery(Document):
    """ History object to retrieve a query to a specific step
    """
    user_id = fields.IntField()  # User link to the history
    query_id = fields.StringField()  # Query link to the history
    message = fields.StringField()  # History message
    status = fields.IntField()  # History status (0: Query step, 1: Waiting for treatment, 2: Output files created)

    @staticmethod
    def get_by_id(history_query_id):
        """ Return a HistoryQuery given its id.

        Args:
            history_query_id:

        Returns:

        """
        try:
            return HistoryQuery.objects.get(pk=str(history_query_id))
        except mongoengine_errors.DoesNotExist as e:
            raise exceptions.DoesNotExist(e.message)
        except Exception as ex:
            raise exceptions.ModelError(ex.message)

    @staticmethod
    def get_all():
        """ Return a list of all HistoryQuery.

        Returns:

        """
        return HistoryQuery.objects().all()

    @staticmethod
    def get_all_by_user_id(user_id):
        """ Return a list of HistoryQuery given their user_id.

        Args:
            user_id:

        Returns:

        """
        return HistoryQuery.objects.filter(user_id=str(user_id))

    def delete_database(self, from_query_admin=False):
        """ Delete function.

        If the deletion come from the admin panel, Delete all the user query linked to this history.

        Args:
            from_query_admin: Variable to avoid circle deletion.
        """
        if not from_query_admin:
            temp_user_queries = TempUserQuery.objects.filter(id=self.query_id)
            for temp_user_query in temp_user_queries:
                temp_user_query.delete_database(from_history=True)
        self.delete()


class TempUserQuery(Document):
    """ Model which define the query associated to a user choice
    """
    query = fields.ReferenceField(DynQuery, blank=True)

    current_position = fields.IntField()  # Current position in the query
    number_of_step = fields.IntField()  # Number of steps
    number_of_query_able_step = fields.IntField()  # Number of query able steps
    # List of steps associated to the query
    list_steps = fields.ListField(fields.ReferenceField(TempUserStep, blank=True), blank=True)
    str_id_file_json = fields.StringField()  # Output JSON file id when it has been created
    str_id_file_xml = fields.StringField()  # Output XML file id when it has been created
    str_id_file_csv = fields.StringField()  # Output CSV file id when it has been created
    last_modified = fields.DateTimeField()  # Last modified timestamp
    history = fields.ReferenceField(HistoryQuery, blank=True)  # History ID linked to the query

    @staticmethod
    def get_by_id(temp_user_query_id):
        """Return a TempUserQuery given its id.

        Args:
            temp_user_query_id:

        Returns:

        """
        try:
            return TempUserQuery.objects.get(pk=str(temp_user_query_id))
        except mongoengine_errors.DoesNotExist as e:
            raise exceptions.DoesNotExist(e.message)
        except Exception as ex:
            raise exceptions.ModelError(ex.message)

    @staticmethod
    def get_by_history_id(history_id):
        """Return a TempUserQuery given its history id.

        Args:
            history_id:

        Returns:

        """
        try:
            return TempUserQuery.objects.get(history=str(history_id))
        except mongoengine_errors.DoesNotExist as e:
            raise exceptions.DoesNotExist(e.message)
        except Exception as ex:
            raise exceptions.ModelError(ex.message)

    @staticmethod
    def get_all():
        """Return all TempUserQuery.

        Returns:

        """
        return TempUserQuery.objects().all()

    def delete_database(self, from_history=False):
        """ Delete function

        Delete each steps of the user query and each part of it.

        Args:
            from_history: If deletion coming from the history, variable to avoid circle deletion.
        """
        for step in self.list_steps:
            try:
                obj_step = TempUserStep.objects.get(id=step.id)
                obj_step.delete_database()
            except:
                pass

        # Delete if to treat
        queries = QueryToTreat.objects.filter(query=self)
        for query in queries:
            query.delete_database()

        # Delete History
        if not from_history:
            history_queries = HistoryQuery.objects.filter(query_id=str(self.id))
            for history_query in history_queries:
                history_query.delete_database(from_query_admin=True)

        # Delete output files
        if self.str_id_file_json != "":
            try:
                temp_file = TempOutputFile.objects.get(id=self.str_id_file_json)
                temp_file.delete_database()
            except DoesNotExist:
                pass
        if self.str_id_file_xml != "":
            try:
                temp_file = TempOutputFile.objects.get(id=self.str_id_file_xml)
                temp_file.delete_database()
            except DoesNotExist:
                pass
        if self.str_id_file_csv != "":
            try:
                temp_file = TempOutputFile.objects.get(id=self.str_id_file_csv)
                temp_file.delete_database()
            except DoesNotExist:
                pass

        # DeleteInfoRedis
        try:
            redis_server = Redis.from_url(REDIS_URL)
            if redis_server.exists("list_ids"):
                try:
                    redis_server.lrem("list_ids", str(self.id), 0)
                except AttributeError:
                    pass
        except ConnectionError:
            pass

        self.delete()

    def initialize_query(self, query_id=None):
        """ Initialise the query used step after step.

        It initialize a query used step after step. If the query name is given, the information
         are loaded from the query define by the admin into the query and his own steps.

        Args:
            query_id: The query name to load.

        Returns:
            The object initialized.
        """
        self.current_position = 0
        self.number_of_step = 0
        self.number_of_query_able_step = 0
        self.list_steps = list()
        self.last_modified = datetime.now()
        self.str_id_file_json = ""
        self.str_id_file_xml = ""
        self.str_id_file_csv = ""

        if query_id is not None:
            self.load_from_query_name_admin(query_id)

    def load_from_query_name_admin(self, query_id):
        """ Load the query from the admin query.

        The information are loaded from the query define by the admin into the query and his
         own steps. It is mainly used by the initialize_query method.


        Args:
            query_id: The query name to load.
        """
        query = DynQuery.objects.get(pk=query_id)
        self.query = query
        self.number_of_step = len(query.steps)

        position = 1
        position_to_show = 1
        full_x_path = ''
        for step in query.steps:
            # Create a new step
            new_step = TempUserStep()
            new_step.initialize_step(step=step)

            # Set the full XPath
            if new_step.step.xpath != ".":
                full_x_path += new_step.step.xpath
            new_step.full_xpath = full_x_path

            # Set and update the positions
            new_step.absolute_position = position
            if new_step.query_able:
                new_step.viewable_position = position_to_show
                position_to_show += 1
                self.number_of_query_able_step += 1
            else:
                new_step.viewable_position = 0

            position += 1

            self.list_steps.append(new_step)

    def save_whole_query(self):
        """ Save the whole query.

        Save the whole query:
        _save the general information,
        _save each steps,
        _save the choices-id_files.
        """

        for step in self.list_steps:
            for choice, list_files in step.dict_choices_id_file.iteritems():
                new_temp_choice_list_file = TempChoiceListFile()

                chunks = [list_files[x:x + 300] for x in xrange(0, len(list_files), 300)]
                list_bucket_temp = list()
                for chunk in chunks:
                    new_bucket = TempBucketIdFiles()
                    new_bucket.list_files = chunk
                    new_bucket.save()
                    list_bucket_temp.append(new_bucket)

                new_temp_choice_list_file.choice = choice
                new_temp_choice_list_file.list_buckets = list_bucket_temp
                new_temp_choice_list_file.save()

                step.list_choices_id_file.append(new_temp_choice_list_file)
            step.save()
        self.save()

    def update_current_position_into_the_db(self):
        """ Update the current position to the object in database.
        """
        self.update(current_position=self.current_position)

    def get_and_load_choices_first_step(self):
        """ Load the first query able step and return it.

        Returns:
             The first query able step
        """
        first_step = self.get_first_query_able_step()

        if first_step is None:
            return None
        else:
            full_projection = "dict_content." + first_step.full_xpath
            if first_step.step.target == 'attribute':
                full_projection += ".@" + first_step.step.value
            elif first_step.step.target == 'name':
                full_projection += ".@title"
            # FIXME: #text was hardcoded, but elements don't have #text if the xml tag does not have any attributes
            # else:
            #     full_projection += ".#text"
            full_projection = possible_projection(full_projection)
            dict_query = {'template': ObjectId(self.query.schema)}

            cursor_xml_data = system_api.execute_query_with_projection(dict_query, full_projection)

            if cursor_xml_data.count() == 0:
                schema_name = Template.objects.get(pk=self.query.schema).display_name
                raise EmptyChoicesFromQuery(schema=schema_name)

            elif first_step.step.data_type == "data":
                first_step.load_choices_and_xml_data(cursor_xml_data=cursor_xml_data)
            else:
                for file_xml_data in cursor_xml_data:
                    first_step.files.append(str(file_xml_data.id))

            return first_step

    def get_and_load_choices_next_step(self):
        """ Get and load the next query able step.

        Returns:
             the next query able step
        """
        current_step = self.get_current_step()
        next_step = self.get_next_query_able_step()
        if next_step is None:
            return None
        if current_step.step.data_type == 'data':
            unique_ids = current_step.get_id_files_from_user_choices()
        else:
            unique_ids = [ObjectId(x) for x in list(set(current_step.files))]

        if next_step.step.data_type == 'data':
            full_projection = 'dict_content.' + next_step.full_xpath
            if next_step.step.target == 'attribute':
                full_projection += ".@" + next_step.step.value
            elif next_step.step.target == 'name':
                full_projection += ".@title"
            # TODO: remove #text hardcoded
            # else:
            #     full_projection += ".#text"
            full_projection = possible_projection(full_projection)
            dict_query = {
                "template": ObjectId(self.query.schema),
                "_id": {"$in": unique_ids}
            }

            cursor_xml_data = system_api.execute_query_with_projection(dict_query, full_projection)

            if cursor_xml_data.count() == 0:
                schema_name = Template.objects.get(id=self.query.schema).display_name
                raise EmptyChoicesFromQuery(schema=schema_name)

            next_step.dict_choices_id_file = dict()
            next_step.load_choices_and_xml_data(cursor_xml_data=cursor_xml_data)
            next_step.update_choices_id_files_to_db()
        else:
            next_step.files = unique_ids
            next_step.update_files_in_db()

        return next_step

    def get_current_step(self):
        """ Return the step corresponding to the current position.

        Returns:
             Current step.
        """
        return self.list_steps[self.current_position]

    def get_first_query_able_step(self):
        """ Return the first query able step. If there is no query able step, it returns None.

        Returns:
            Return None if there is no query able step, or the first query able step.
        """
        self.current_position = 0
        if self.is_query_able(0):
            return self.list_steps[0]
        return self.get_next_query_able_step()

    def get_last_query_able_step(self):
        """ Return the last query able step. If there is no query able step, it returns None.

        Returns:
             Return None if there is no query able step, or the last query able step.
        """
        for step in reversed(self.list_steps):
            if step.query_able is True:
                return step
        return None

    def get_next_query_able_step(self):
        """ Return the next query able step. If there is no query able step, it returns None.

        Returns:
             Return None if there is no next query able step, or the next query able step.
        """
        current_viewable_position = self.list_steps[self.current_position].viewable_position
        if current_viewable_position == self.number_of_query_able_step:
            return None
        else:
            next_viewable_position = current_viewable_position + 1
            self.current_position += 1
            while self.list_steps[self.current_position].viewable_position \
                    != next_viewable_position:
                self.current_position += 1
            return self.list_steps[self.current_position]

    def get_previous_query_able_step(self):
        """ Return the previous query able step. If there is no query able step, it returns None.

        Returns:
             Return None if there is no previous query able step, or the previous query able step.
        """
        current_viewable_position = self.list_steps[self.current_position].viewable_position
        if current_viewable_position == 1:
            return None
        else:
            self.list_steps[self.current_position].clean_step()
            previous_viewable_position = current_viewable_position - 1
            self.current_position -= 1
            while self.list_steps[self.current_position].viewable_position != \
                    previous_viewable_position:
                self.list_steps[self.current_position].clean_step()
                self.current_position -= 1

            self.list_steps[self.current_position].update(choices=list())
            return self.list_steps[self.current_position]

    def get_previous_choices(self):
        """ Get the formatted the previous user choices.

        Return the previous user choices separated by "/".

        Returns:
             Return the formatted the previous user choices separated by "/".
        """
        prev_choices = list()
        for position in range(0, self.current_position + 1):
            if self.is_query_able(position):
                prev_choices.append(str(self.list_steps[position].choices))
        return " / ".join(prev_choices)

    def get_ids_files_last_step(self):
        """ Get the minimal list of files where data can be picked in. If there is no query able step, ie no choices,
        all the files linked to the query's schema will be returned.

        Returns:
             Minimal list of files from the database linked to the user choices.
        """
        last_query_able_step = self.get_last_query_able_step()
        if last_query_able_step is None:
            xml_data_list = system_api.get_all_by_template(ObjectId(self.query.schema))
            list_ids = list()
            for xml_data in xml_data_list:
                list_ids.append(xml_data.id)
            return list_ids

        if last_query_able_step.step.data_type == 'data':
            id_current_files = []
            for choice_id_file in last_query_able_step.list_choices_id_file:
                if choice_id_file.choice in last_query_able_step.choices:
                    for bucket in choice_id_file.list_buckets:
                        id_current_files.extend(bucket.list_files)
        else:
            id_current_files = last_query_able_step.files

        return [ObjectId(x) for x in list(set(id_current_files))]

    def is_query_able(self, position=None):
        """ Return if a step is query able.

        It returns True if the step is query able, False if the step is not query able. The
        tested step is the first one if the position is not given,
        or the step corresponding to the given position.

        Args:
            position: The position's step designing the tested step.

        Returns:
            Return True if the step is query able, False if the step is not query able.
        """
        if position is None:
            return self.list_steps[self.current_position].query_able
        else:
            return self.list_steps[position].query_able

    def is_current_first_step(self):
        """ Return if the current step is the first step.

        It returns True if the step is the first step, False if the step is not the first one.

        Returns:
            Return True if the step is query able, False if the step is not query able.
        """
        return self.get_current_step().viewable_position == 1

    def update_time(self):
        """ Update the query name.
        """
        self.update(last_modified=datetime.now())

    def save_to_history(self, user_id, history_id=None):
        """ Create or update the history query.

        Args:
            user_id: User id.
            history_id: History query associated. If none, the history has to be created.

        Returns:
            History id.
        """
        if history_id is None:
            history = HistoryQuery()
            history.query_id = str(self.id)
            history.status = 0
            history.user_id = user_id
            history.message = "Step " + \
                              str(self.get_current_step().viewable_position) + \
                              ": " + \
                              self.get_current_step().step.name
            history_id = history.save()
            self.update(history=history_id)
        else:
            history = HistoryQuery.objects.get(id=history_id)
            history.update(message="Step "
                                   + str(self.get_current_step().viewable_position)
                                   + ": "
                                   + self.get_current_step().step.name
                           )
        return history.id

    def update_message(self, _message):
        """ Update History message

        Args:
            _message: Message to be updated
        """
        TempUserQuery.objects.get(id=self.id).history.update(message=_message)

    def handle_history(self):
        """ Get previous choice and return step
        """
        choices = self.get_previous_choices()

        if choices[:-1] is not []:
            step = self.get_current_step()
            step.choices = []
            step.update(choices=[])
        else:
            raise ModelError("Previous choice does not exist")

        return step

    def create_files(self, dict_data, map_keys, list_headers, list_file_xml, list_file_json,
                     list_file_csv, depth, max_depth):
        """ Create the output file

        Args:
            dict_data: Tree of data.
            map_keys: Map a hash key to a key. The key is a dictionary representing a node.
            list_headers: List of headers parts common to all current element.
            list_file_xml: Different part of xml file.
            list_file_json: Different part of json file.
            list_file_csv: Different part of csv file.
            depth: Current depth in the tree.
            max_depth: Maximum depth in the tree.

        Returns:
            Output files.
        """

        if depth == max_depth:  # Leaf level
            if dict_data:
                list_keys = dict_data.keys()

                # Sort data if possible
                is_timestamp_present = True
                for hash_data in list_keys:
                    if "@timestamp" not in map_keys[hash_data].keys():
                        is_timestamp_present = False
                        break
                if is_timestamp_present:
                    list_keys = sorted(dict_data.keys(),
                                       key=lambda x: map_keys[x]['@timestamp'])

                # Keeping only the 10 000 last records
                if self.query.is_limited:
                    if self.query.number_records is not None:
                        list_keys = list_keys[-self.query.number_records:]

                if len(dict_data) > 1:
                    common_keys, list_specific_header = get_common_key_and_specific_header(
                        list_keys, map_keys)
                else:
                    common_keys = list()
                    list_specific_header = list_headers[:]  # Copy by slicing

                ddict_title_data = get_title_data_leaf(list_keys, map_keys)
                list_specific_header = list_headers + list_specific_header

                print_bloc(list_specific_header, ddict_title_data, common_keys,
                           list_file_xml,
                           list_file_json, list_file_csv)

            return
        else:  # Header level
            for key_hash, value in dict_data.iteritems():
                dict_key = map_keys[key_hash]
                list_header_actual_key = get_header_parents(dict_key)
                self.create_files(
                    dict_data=dict_data[key_hash],
                    map_keys=map_keys,
                    list_headers=list_headers + list_header_actual_key,
                    list_file_xml=list_file_xml,
                    list_file_json=list_file_json,
                    list_file_csv=list_file_csv,
                    depth=depth + 1,
                    max_depth=max_depth
                )
        return

    def init_data_result(self):
        """ Initialize the variables to create the files. Create the files and save them into the database.
        """
        translate_dt = maketrans("-:.T", '    ')  # Translate object for element date time
        translate_bounds = maketrans("-:", '  ')  # Translate object for bounds date time
        max_depth = self.number_of_step

        # Create the query
        ids_xml_data = self.get_ids_files_last_step()
        nb_files_total = len(ids_xml_data)
        first_step_xpath = self.list_steps[0].step.xpath.split("*")[0]
        full_projection = possible_projection("dict_content." + first_step_xpath)

        dict_query = {
            "template": ObjectId(self.query.schema),
            "_id": {
                "$in": ids_xml_data
            }
        }

        xml_data = [
            data.to_mongo()['dict_content']
            for data in system_api.execute_query_with_projection(dict_query, full_projection)
        ]

        depth = 0
        list_keys_xpath = []
        dict_data = dict()
        map_key = dict()
        list_nb_file_treated = list()
        list_nb_file_treated.append(1.0)

        # Get the filtered elements form the database.
        self.explore_elements(list(xml_data), list_keys_xpath, depth, max_depth, dict_data,
                              translate_dt, translate_bounds, map_key, list_nb_file_treated, nb_files_total)

        self.update_message("Creating output files.")

        list_step_dot_only = [x for x in self.list_steps if x.step.xpath == "."]
        max_depth -= len(list_step_dot_only) + 1

        # Create the output files
        depth = 0
        list_file_xml = list()
        list_file_json = list()
        list_file_csv = list()
        list_headers = list()

        list_file_json.append("{\r\n\t\"Data\": {\r\n\t\t\"Item\": [\r\n")
        list_file_xml.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n<Data>\r\n")

        self.create_files(dict_data, map_key, list_headers, list_file_xml, list_file_json,
                          list_file_csv, depth, max_depth)

        file_csv = ''.join(list_file_csv)
        list_file_xml.append('</Data>')
        file_xml = ''.join(list_file_xml)
        file_json = ''.join(list_file_json)
        file_json = file_json[:-6] + file_json[-6:].replace(',', '')
        file_json += "\t\t]\r\n\t}\r\n}"

        self.update(str_id_file_json=TempOutputFile().save_file(str_file=file_json, type_file="json"))
        self.update(str_id_file_csv=TempOutputFile().save_file(str_file=file_csv, type_file="csv"))
        self.update(str_id_file_xml=TempOutputFile().save_file(str_file=file_xml, type_file="xml"))

        return

    def explore_elements(self, list_elements, list_keys_xpath, depth, max_depth, dict_to_add,
                         translate_dt, translate_bounds, map_key, list_nb_file_treated, nb_file_total):
        """ Explore the elements and filter them from an input file.

        Args:
            list_elements: List of current element.
            list_keys_xpath: List of sliced xpath.
            depth: Current depth in the element.
            max_depth: Maximum element depth.
            dict_to_add: Result dictionary.
            translate_dt: Translate python object for element date.
            translate_bounds: Translate python object for bounds date.
            map_key: Map between hash and dictionary.
            list_nb_file_treated: List used for the advancement percentage. Only the first element is used.
            nb_file_total:  Number of total files.
        """
        if depth == 1:
            self.update_message(
                "Gathering all data. - "
                + str("%.2f" % ((list_nb_file_treated[0] / nb_file_total) * 100))
                + " %"
            )
            list_nb_file_treated[0] += 1

        if depth == max_depth:
            return
        if not list_keys_xpath:
            # Create the list of xpath element for the current depth
            list_keys_xpath = [x for x in self.list_steps[depth].step.xpath.split(".") if x]
        element_title = list_keys_xpath[-1]

        while list_keys_xpath:
            key_xpath = list_keys_xpath.pop(0)
            if key_xpath == "*":
                # Solve start element
                set_possibilities = explore_star(list_elements)
                for key_possible in set_possibilities:
                    list_key_param_star = [key_possible]
                    if len(list_keys_xpath) != 0:
                        list_key_param_star.extend(list_keys_xpath)
                    self.explore_elements(
                        list_elements=list_elements,
                        list_keys_xpath=list_key_param_star,
                        depth=depth,
                        max_depth=max_depth,
                        dict_to_add=dict_to_add,
                        translate_dt=translate_dt,
                        translate_bounds=translate_bounds,
                        map_key=map_key,
                        list_nb_file_treated=list_nb_file_treated,
                        nb_file_total=nb_file_total,
                    )
                return
            else:
                # Go across all elements
                list_elements = [x.get(key_xpath) for x in list_elements if x.get(key_xpath)]

        # Flat the elements if an element i composed by a a list of elements
        if list_elements:
            if isinstance(list_elements[0], list):
                list_elements = flat_list(list_elements)

        # Apply the choices to the list of elements
        depth, list_elements_cleaned = self.apply_user_choices(list_elements, depth, max_depth,
                                                               element_title,
                                                               translate_dt, translate_bounds)

        if not list_elements_cleaned:
            return
        dict_to_str = hash
        for element in list_elements_cleaned:
            dict_key_output_master = get_general_key_output_dictionary(
                element, element_title)
            str_key_output_spec = dict_to_str(frozenset(dict_key_output_master.items()))

            if str_key_output_spec not in dict_to_add:
                map_key[str_key_output_spec] = dict_key_output_master
                dict_to_add[str_key_output_spec] = {}

            self.explore_elements(
                list_elements=[element],
                list_keys_xpath=[],
                depth=depth + 1,
                max_depth=max_depth,
                dict_to_add=dict_to_add[str_key_output_spec],
                translate_dt=translate_dt,
                translate_bounds=translate_bounds,
                map_key=map_key,
                list_nb_file_treated=list_nb_file_treated,
                nb_file_total=nb_file_total,
            )
        return

    def apply_user_choices(self, list_elements, depth, max_depth, element_title, translate_dt, translate_bounds):
        """ Apply user choices to a list of elements.

        Args:
            list_elements: List of elements to be filtered.
            depth: Current depth.
            max_depth: Maximum Depth.
            element_title: Current element title
            translate_dt: Translate python object for the element date time.
            translate_bounds: Translate python object for the bounds.

        Returns:
            depthm list cleaned elements
        """
        current_step = self.list_steps[depth]
        list_cleaned_elements = current_step.choose_filter_choice(
            list_elements=list_elements,
            element_title=element_title,
            translate_dt=translate_dt,
            translate_bounds=translate_bounds
        )

        while depth + 1 != max_depth and self.list_steps[depth + 1].step.xpath == ".":
            depth += 1
            current_step = self.list_steps[depth]
            list_cleaned_elements = current_step.choose_filter_choice(
                list_elements=list_cleaned_elements,  # FIXME: was list_elements (would ignore the first filter)
                element_title=element_title,
                translate_dt=translate_dt,
                translate_bounds=translate_bounds
            )
        return depth, list_cleaned_elements

    def add_query_to_waiting_line(self):
        """ Add query to the waiting line when the user end his choices.
        """
        query_to_treat = QueryToTreat()
        query_to_treat.query = self
        query_to_treat.save()

    def create_outputs_file(self):
        """ Create the output files
        """
        try:
            self.init_data_result()
        except Exception, e:
            current_step = self.get_current_step()
            log_file = LogFile(application="Custom Queries",
                               message="The step #"
                                       + str(current_step.absolute_position)
                                       + ": " + str(current_step.step.name)
                                       + " is not a valid timestamp."
                                       + str(e.message),
                               additionalInformation={'query_name': self.query_name,
                                                      'Status': "Result builder"},
                               timestamp=datetime.now())
            log_file_api.upsert(log_file)
            self.update_message("Error during result files creation.")
            return

        self.update_message("Result files created.")
        # Get the user to send him an email
        try:
            history = HistoryQuery.objects.get(query_id=str(self.id))
            user = User.objects.get(id=history.user_id)
            send_mail_query_end(query=self, user=user)
        except:
            pass
        self.update_status(2)

    def update_status(self, new_status):
        TempUserQuery.objects.get(id=self.id).history.update(status=new_status)


class QueryToTreat(Document):
    """ List of query waiting to be treated.
    """
    query = fields.ReferenceField(TempUserQuery, blank=True)

    @staticmethod
    def get_all():
        """ Return all QueryToTreat.

        Returns:

        """
        return QueryToTreat.objects().all()

    def delete_database(self):
        """ Delete function.
        """
        self.delete()



""" Initialize permissions for core custom queries app
"""
from django.contrib.auth.models import Group, Permission
import core_main_app.permissions.rights as main_rights
import core_custom_queries_app.permissions.rights as custom_queries_rights


def init_permissions():
    """ Initialization of groups and permissions.

    Returns:

    """
    try:
        # Get or Create the default group
        default_group, created = Group.objects.get_or_create(name=main_rights.default_group)

        # Get custom queries permissions
        custom_queries_access_perm = Permission.objects.get(codename=custom_queries_rights.custom_queries_access)

        # Add permissions to default group
        default_group.permissions.add(custom_queries_access_perm)

    except Exception, e:
        print('ERROR : Impossible to init the permissions : ' + e.message)



"""
Describe all the views used by custom_queries
"""
from datetime import datetime

from bson import ObjectId
from django.contrib import messages
from django.core.urlresolvers import reverse
from django.forms import formset_factory
from django.shortcuts import redirect
from django.urls import reverse_lazy
from mongoengine import ValidationError
from redis import Redis, ConnectionError

import core_main_app.utils.decorators as decorators
from core_custom_queries_app.components.dyn_query import api as dyn_query_api
from core_custom_queries_app.components.history_query import api as history_query_api
from core_custom_queries_app.components.log_file import api as log_file_api
from core_custom_queries_app.components.log_file.models import LogFile
from core_custom_queries_app.components.temp_output_file import api as temp_output_file_api
from core_custom_queries_app.components.temp_user_query import api as temp_user_query_api
from core_custom_queries_app.components.temp_user_step.models import TempUserStep
from core_custom_queries_app.models import TempUserQuery
from core_custom_queries_app.permissions import rights
from core_custom_queries_app.views.user.forms import FormChooseQuery, FormRadio, FormCheck, FormDateTime, FormResult, \
    FormHistory
from core_main_app.commons.exceptions import DoesNotExist
from core_main_app.components.template_version_manager import api as template_version_manager_api
from core_main_app.utils.file import get_file_http_response
from core_main_app.utils.rendering import render
from qdr.settings import REDIS_URL


@decorators.permission_required(content_type=rights.custom_queries_content_type,
                                permission=rights.custom_queries_access, login_url=reverse_lazy("core_main_app_login"))
def choose_query(request):
    """
    Select a query from the one into the database to be able to use it.
    Used by the user.
    If there is no query into the database, an error message is shown to the user.
    If no query has been selected, an error message is shown to the user.
    Send to the query_step if a query has been selected.
    If history queries are available, they are show with the following option:
    _History deletion
    _History recovery

    Args:
        request:

    Returns:

    """
    formset = None
    formset_history = None
    history_queries = None
    # Come back from the form with the query selected
    if request.method == 'POST':
        # Get the forms variable
        query_id = request.POST.get('query', None)
        to_recover = request.POST.get('to_recover', None)
        number_forms = request.POST.get('form-TOTAL_FORMS', None)

        # if a variable is none, come back with an error message
        if query_id is None and to_recover is None and number_forms is None:
            messages.add_message(request, messages.INFO, "You have to select a query.")
        else:
            if query_id is not None:
                return redirect(reverse("core_custom_queries_app_steps", kwargs={'query_id': query_id}))
            # History selected
            elif to_recover is not None and to_recover != '':
                return redirect(reverse("core_custom_queries_app_recover_steps", kwargs={'history_id': to_recover}))
            # Delete history query
            else:
                number_forms = int(number_forms)
                formset_history = formset_factory(FormHistory, extra=number_forms,
                                                  can_delete=True)
                formset = formset_history(request.POST, request.FILES)
                if formset.is_valid():
                    if len(formset.deleted_forms) == 0:
                        messages.add_message(request, messages.ERROR,
                                             "Please select a query first.")
                    for form in formset.deleted_forms:
                        data = form.cleaned_data
                        if 'DELETE' in data:
                            id_history = ObjectId(data['id_history'])
                            history = history_query_api.get_by_id(id_history)
                            temp_query_id = history.query_id
                            query = temp_user_query_api.get_by_id(temp_query_id)
                            query.delete_database()
                    messages.add_message(request, messages.SUCCESS,
                                         "Queries successfully deleted.")

    # Create the query form from the ones into the database
    id_list = list()
    # Get the available templates
    available_template_version_managers = template_version_manager_api.get_active_global_version_manager()
    for available_template_version_manager in available_template_version_managers:
        for available_template_id in available_template_version_manager.versions:
            id_list.append(str(available_template_id))

    # Get the queries available for the current user
    if request.user.is_staff:
        # FIXME: call specialized api
        queries = dyn_query_api.get_all().filter(schema__in=id_list).values_list('pk', 'name')
    else:
        # FIXME: call specialized api
        queries = dyn_query_api.get_all().filter(schema__in=id_list).filter(group='user').values_list('pk', 'name')

    number_queries = len(queries)

    form = None
    if number_queries != 0:
        form = FormChooseQuery(query=queries)
    else:
        # If no query available, set an error message
        message = "No query available."
        messages.add_message(request, messages.ERROR, message)

    try:
        # If histories exist load them
        history_queries = history_query_api.get_all_by_user_id(str(request.user.id))
        number_history = len(history_queries)
    except Exception:
        number_history = 0

    if number_history > 0:
        # Create the formset with the histories values
        formset_history = formset_factory(FormHistory, extra=number_history, can_delete=True)
        data = dict()
        data['form-TOTAL_FORMS'] = number_history
        data['form-INITIAL_FORMS'] = number_history
        data['form-MAX_NUM_FORMS'] = number_history
        position = 0
        for history_query in history_queries:
            try:
                query_id = str(history_query.query_id)
                data['form-' + str(position) + '-id_history'] = str(history_query.id)
                query = temp_user_query_api.get_by_id(query_id)
                data['form-' + str(position) + '-query_name'] = query.query.name
                data['form-' + str(position) + '-query_last_modified'] = query.last_modified

                h_message = history_query.message
                if h_message == "Pending.":
                    redis_server = Redis.from_url(REDIS_URL)
                    try:
                        if redis_server.exists("current_id"):
                            current_id = redis_server.get("current_id")
                            if current_id == query_id:
                                h_message = "Pending: The file creation will begin shortly."

                        if redis_server.exists("list_ids"):
                            list_id = redis_server.lrange('list_ids', 0, -1)
                            try:
                                position_id = list_id.index(query_id) + 1
                                h_message = "Pending - Waiting list: " + str(position_id) + "/" + str(len(list_id))
                            except ValueError:
                                pass

                    except ConnectionError, e:
                        log_file = LogFile(application="Custom Queries",
                                           message="Redis not reachable, is it running?",
                                           additionalInformation={'message': e.message},
                                           timestamp=datetime.now())
                        log_file_api.upsert(log_file)

                data['form-' + str(position) + '-query_message'] = h_message
                data['form-' + str(position) + '-status'] = history_query.status
                data['form-' + str(position) + '-DELETE'] = ''
                position += 1
            except DoesNotExist:
                pass
        formset = formset_history(data)

    context = {'number_queries': number_queries,
               'number_history': number_history,
               'form': form,
               'formset': formset,
               'formset_history': formset_history}

    assets = {
        "js": [
            {
                "path": 'core_custom_queries_app/user/js/custom_queries.js',
                "is_raw": False
            },
        ],
    }

    return render(request,
                  'core_custom_queries_app/user/select_query.html',
                  assets=assets,
                  context=context)


@decorators.permission_required(content_type=rights.custom_queries_content_type,
                                permission=rights.custom_queries_access, login_url=reverse_lazy("core_main_app_login"))
def query_steps(request, query_id):
    """
    Manage each step of the query.

    Create the query step in function of the step type. Validate the user choices and store each step to the query
    history. The last step validation store the query to the database to explore the files and get the data.

    Args:
        request:
        query_id:

    Returns:

    """
    user_query = TempUserQuery()
    user_step = TempUserStep()
    first_step = False
    previous_step = False
    next_step = False
    history_id = None

    # Validate the user step
    if request.method == 'POST':
        # Get the form's value
        user_query_id = request.POST.get('user_query', None)
        step_back = request.POST.get('step_back', None)
        history_id = request.POST.get('history_id', None)

        # Handle value transportation problems.
        if user_query is None or step_back is None or history_id is None:
            messages.add_message(request, messages.ERROR,
                                 "A problem occurred during your information's transportation."
                                 " Please do it again.")
            return redirect(reverse("custom_queries_queryChoose"))
        try:
            # Load the query
            user_query = temp_user_query_api.get_by_id(user_query_id)
        except DoesNotExist:
            # The query doesn't exist anymore.
            messages.add_message(request, messages.ERROR,
                                 "A problem occurred during your information's transportation."
                                 " Please do it again.")

            return redirect(reverse("custom_queries_queryChoose"))

        # Get the current step information
        user_step = user_query.get_current_step()
        form_type = user_step.form_type
        # If the user want to go back to the previous step
        if step_back == 'True':
            previous_step = True
        else:
            error = False
            # The step is a a single choice step
            if form_type == 'radio':
                choices = request.POST.get('choices', None)
                if choices:
                    user_step.choices.append(choices)
                else:
                    messages.add_message(request, messages.WARNING, 'You have to choose one.')
                    error = True
            # The step is a a multiple choices step
            elif form_type == 'check':
                list_choices = request.POST.getlist('choices')
                # Get the step restrictions and validate the choices against the restrictions
                min_range = user_step.step.data_min_range
                if min_range != '' and min_range:
                    min_range = int(min_range)
                    max_range = user_step.step.data_max_range
                    data_infinity = user_step.step.data_infinity
                    if max_range != '' and max_range:
                        max_range = int(max_range)
                        if max_range < len(list_choices) or len(list_choices) < min_range:
                            error = True
                            messages.add_message(request, messages.WARNING,
                                                 'You have to choose between '
                                                 + str(min_range)
                                                 + ' and ' + str(max_range)
                                                 + ' choice.')
                    else:
                        if len(list_choices) < min_range:
                            error = True
                            messages.add_message(request, messages.WARNING,
                                                 'You have to choose at least one.')
                if not error:
                    user_step.choices = list_choices
            # The step is a a date step
            else:
                time_from = request.POST['timeFrom']
                time_to = request.POST['timeTo']
                if time_from == "" or time_to == "":
                    error = True
                    messages.add_message(request, messages.WARNING, 'You have to fill both date.')
                else:
                    try:
                        time_from = datetime.strptime(time_from, "%Y-%m-%d %H:%M")
                        time_to = datetime.strptime(time_to, "%Y-%m-%d %H:%M")

                        if time_from > time_to:
                            error = True
                            messages.add_message(request, messages.WARNING,
                                                 'Date \"From\" has to be before date \"To\"')

                        date_range = user_step.step.date_range
                        if date_range and date_range != '':
                            date_range = int(date_range)
                            if (time_to - time_from).days > date_range:
                                error = True
                                messages.add_message(request, messages.WARNING,
                                                     'The maximum time range you can query in is '
                                                     + str(date_range)
                                                     + ' day(s).')
                    except Exception, e:
                        error = True
                        messages.add_message(request, messages.ERROR, e)
                if not error:
                    user_step.choices = (request.POST['timeFrom'], request.POST['timeTo'])
            if not error:
                user_step.update_choices_to_db()
                next_step = True
    # Select and load the current step
    else:
        # Initialize the query
        user_query = TempUserQuery()
        user_query.initialize_query(query_id=query_id)
        first_step = True

    try:
        # Load the current step
        if first_step is True:
            # Load the first step
            user_step = user_query.get_and_load_choices_first_step()
            user_query.save_whole_query()
        elif previous_step is True:
            # Load the previous step
            user_step = user_query.get_previous_query_able_step()
            user_query.update_current_position_into_the_db()
            user_step.transform_choices_files_list_to_dict()
        elif next_step is True:
            # Load the next step
            user_step = user_query.get_and_load_choices_next_step()
            user_query.update_current_position_into_the_db()
        else:
            # Error validation
            user_step.transform_choices_files_list_to_dict()
        history_id = user_query.save_to_history(user_id=str(request.user.id), history_id=history_id)
    except Exception, e:
        # Create the log file from the error
        log_file = LogFile(application="Custom Queries",
                           message=e.message,
                           additionalInformation={'query_name': user_query.query.name,
                                                  'step_name': user_query.list_steps[user_query.current_position].step.name,
                                                  'user_choices': user_query.get_previous_choices()},
                           timestamp=datetime.now())
        log_file_api.upsert(log_file)
        try:
            user_query.update_message("Error during query step.")
        except ValidationError:
            pass

        messages.add_message(request, messages.ERROR,
                             "An internal problem occurred, the administrator has been notified.")
        return redirect(reverse("core_custom_queries_app_index"))

    if user_step is None:
        # Go to result page
        user_query.add_query_to_waiting_line()
        user_query.update_message("Pending.")
        user_query.update_status(1)
        messages.add_message(request, messages.SUCCESS,
                             "The output file will be created soon.")
        return redirect(reverse("core_custom_queries_app_index"))
    else:
        context = _create_step_context(user_query, user_step, history_id)

        assets = {
            "js": [
                {
                    "path": 'core_custom_queries_app/user/js/custom_queries.js',
                    "is_raw": False
                },
                {
                    "path": 'core_custom_queries_app/user/js/custom_queries_step.js',
                    "is_raw": False
                },
                {
                    "path": "core_custom_queries_app/libs/bootstrap-datetimepicker/js/bootstrap-datetimepicker.js",
                    "is_raw": False
                },
            ],
            'css': ['core_custom_queries_app/libs/bootstrap-datetimepicker/css/bootstrap-datetimepicker.min.css']
        }

        return render(request,
                      'core_custom_queries_app/user/query_step.html',
                      assets=assets,
                      context=context)


# FIXME: refactor code duplicated with query_steps
@decorators.permission_required(content_type=rights.custom_queries_content_type,
                                permission=rights.custom_queries_access, login_url=reverse_lazy("core_main_app_login"))
def recover_query_steps(request, history_id):
    """

    """
    user_step = None
    previous_step = False
    next_step = False
    history_step = False

    # Validate the user step
    if request.method == 'POST':
        # Get the form's value
        user_query_id = request.POST.get('user_query', None)
        step_back = request.POST.get('step_back', None)
        try:
            # Load the query
            user_query = temp_user_query_api.get_by_id(user_query_id)
        except DoesNotExist:
            # The query doesn't exist anymore.
            messages.add_message(request, messages.ERROR,
                                 "A problem occurred during your information's transportation."
                                 " Please do it again.")

            return redirect(reverse("core_custom_queries_app_index"))

        # Get the current step information
        user_step = user_query.get_current_step()
        form_type = user_step.form_type
        # If the user want to go back to the previous step
        if step_back == 'True':
            previous_step = True
        else:
            error = False
            # The step is a a single choice step
            if form_type == 'radio':
                choices = request.POST.get('choices', None)
                if choices:
                    user_step.choices.append(choices)
                else:
                    messages.add_message(request, messages.WARNING, 'You have to choose one.')
                    error = True
            # The step is a a multiple choices step
            elif form_type == 'check':
                list_choices = request.POST.getlist('choices')
                # Get the step restrictions and validate the choices against the restrictions
                min_range = user_step.step.data_min_range
                if min_range != '' and min_range:
                    min_range = int(min_range)
                    max_range = user_step.step.data_max_range
                    # data_infinity = user_step.step.data_infinity
                    if max_range != '' and max_range:
                        max_range = int(max_range)
                        if max_range < len(list_choices) or len(list_choices) < min_range:
                            error = True
                            messages.add_message(request, messages.WARNING,
                                                 'You have to choose between '
                                                 + str(min_range)
                                                 + ' and ' + str(max_range)
                                                 + ' choice.')
                    else:
                        if len(list_choices) < min_range:
                            error = True
                            messages.add_message(request, messages.WARNING,
                                                 'You have to choose at least one.')
                if not error:
                    user_step.choices = list_choices
            # The step is a a date step
            else:
                time_from = request.POST['timeFrom']
                time_to = request.POST['timeTo']
                if time_from == "" or time_to == "":
                    error = True
                    messages.add_message(request, messages.WARNING, 'You have to fill both date.')
                else:
                    try:
                        time_from = datetime.strptime(time_from, "%Y-%m-%d %H:%M")
                        time_to = datetime.strptime(time_to, "%Y-%m-%d %H:%M")

                        if time_from > time_to:
                            error = True
                            messages.add_message(request, messages.WARNING,
                                                 'Date From has to be before date To')

                        date_range = user_step.step.date_range
                        if date_range and date_range != '':
                            date_range = int(date_range)
                            if (time_to - time_from).days > date_range:
                                error = True
                                messages.add_message(request, messages.WARNING,
                                                     'The maximum time range you can query in is '
                                                     + str(date_range)
                                                     + ' day(s).')
                    except Exception, e:
                        error = True
                        messages.add_message(request, messages.ERROR, e)
                if not error:
                    user_step.choices = (request.POST['timeFrom'], request.POST['timeTo'])
            if not error:
                user_step.update_choices_to_db()
                next_step = True
    # Select and load the current step
    else:
        # Load history queries
        try:
            history_query = history_query_api.get_by_id(history_id)
            user_query = temp_user_query_api.get_by_id(history_query.query_id)
        except DoesNotExist:
            messages.add_message(request, messages.ERROR, "A problem occurred during the redirection.")
            return redirect(reverse("core_custom_queries_app_index"))
        history_step = True
    try:        # Load the current step
        if previous_step is True:
            # Load the previous step
            user_step = user_query.get_previous_query_able_step()
            user_query.update_current_position_into_the_db()
            user_step.transform_choices_files_list_to_dict()
            history_id = user_query.save_to_history(user_id=request.user.id, history_id=history_id)
        elif next_step is True:
            # Load the next step
            user_step = user_query.get_and_load_choices_next_step()
            user_query.update_current_position_into_the_db()
            history_id = user_query.save_to_history(user_id=request.user.id, history_id=history_id)
        elif history_step is True:
            if user_query.history.message == "Error during output file creation.":
                # An error occurred
                messages.add_message(request, messages.ERROR,
                                     "An internal problem occurred during the output file creation. "
                                     "The administrator has been noticed.")
                return redirect(reverse("core_custom_queries_app_index"))
            else:
                # Load the history query
                user_step = user_query.handle_history()
                user_step.transform_choices_files_list_to_dict()
    except Exception, e:
        # Create the log file from the error
        log_file = LogFile(application="Custom Queries",
                           message=e.message,
                           additionalInformation={'query_name': user_query.query.name,
                                                  'step_name': user_query.list_steps[
                                                      user_query.current_position].step.name,
                                                  'user_choices': user_query.get_previous_choices()},
                           timestamp=datetime.now())
        log_file_api.upsert(log_file)
        try:
            user_query.update_message("Error during query step.")
        except ValidationError:
            pass

        messages.add_message(request, messages.ERROR,
                             "An internal problem occurred, the administrator has been notified.")
        return redirect(reverse("core_custom_queries_app_index"))

    if user_step is None:
        # Go to result page
        user_query.add_query_to_waiting_line()
        user_query.update_message("Pending.")
        user_query.update_status(1)
        messages.add_message(request, messages.SUCCESS,
                             "The output file will be created soon.")
        return redirect(reverse("core_custom_queries_app_index"))
    else:
        context = _create_step_context(user_query, user_step, history_id)

        assets = {
            "js": [
                {
                    "path": 'core_custom_queries_app/user/js/custom_queries.js',
                    "is_raw": False
                },
                {
                    "path": 'core_custom_queries_app/user/js/custom_queries_step.js',
                    "is_raw": False
                },
                {
                    "path": "core_custom_queries_app/libs/bootstrap-datetimepicker/js/bootstrap-datetimepicker.js",
                    "is_raw": False
                },
            ],
            'css': ['core_custom_queries_app/libs/bootstrap-datetimepicker/css/bootstrap-datetimepicker.min.css']
        }

        # for the rendering
        return render(request,
                      'core_custom_queries_app/user/query_step.html',
                      assets=assets,
                      context=context)


def _create_step_context(user_query, user_step, history_id):
    """

    Args:
        user_query:
        user_step:
        history_id:

    Returns:

    """
    user_query.update_time()
    # Next step query data
    query_name = user_query.query.name
    position = user_step.viewable_position
    form_type = user_step.form_type

    # Variable use by the templates
    position_to_show = user_step.viewable_position
    nb_step_to_show = user_query.number_of_query_able_step
    user_query_id = user_query.id
    first_step = user_query.is_current_first_step()

    # Form creation
    if user_step.step.data_type == 'data':
        choices = user_step.get_choices_as_tuple()
        step_name = user_step.step.name
        data_infinity = user_step.step.data_infinity
        data_min_range = user_step.step.data_min_range
        data_max_range = user_step.step.data_max_range
        if form_type == 'radio':
            form = FormRadio(choices=choices, queryName=query_name, stepName=step_name,
                             position=position, user_query=user_query_id, data_type=form_type,
                             step_back=False, history_id=history_id)
        else:
            form = FormCheck(choices=choices, queryName=query_name, stepName=step_name,
                             position=position, user_query=user_query_id,
                             data_min_range=data_min_range, data_max_range=data_max_range,
                             data_infinity=data_infinity, data_type=form_type, step_back=False,
                             history_id=history_id)
    # Next step query date
    else:
        date_range = user_step.step.date_range
        form = FormDateTime(queryName=query_name, position=position, user_query=user_query_id,
                            date_range=date_range, data_type=form_type, step_back=False,
                            history_id=history_id)

    context = {
        'query_name': query_name,
        'position_to_show': position_to_show,
        'nb_step_to_show': nb_step_to_show,
        'first_step': first_step,
        'form': form
    }

    return context


@decorators.permission_required(content_type=rights.custom_queries_content_type,
                                permission=rights.custom_queries_access, login_url=reverse_lazy("core_main_app_login"))
def output_files(request, history_id):
    """ Display output files view

    Args:
        request:
        history_id:

    Returns:

    """
    try:
        user_query = temp_user_query_api.get_by_history_id(history_id)
        # Load the step from the history
        if user_query.history.status == 2:
            # If the file has been created, load the files, and render the download file
            id_file_json = user_query.str_id_file_json
            id_file_csv = user_query.str_id_file_csv
            id_file_xml = user_query.str_id_file_xml
            query_name = user_query.query.name
            preview_json = temp_output_file_api.load_file(id_file_json)[:2000]
            preview_csv = temp_output_file_api.load_file(id_file_csv)[:2000]
            if len(preview_csv) == 0:
                preview_csv = 'The file is empty.'
            preview_xml = temp_output_file_api.load_file(id_file_xml)[:2000]
            form = FormResult(
                queryName=query_name,
                preview_file_csv=preview_csv,
                preview_file_json=preview_json,
                preview_file_xml=preview_xml,
                hidden_id_file_csv=id_file_csv,
                hidden_id_file_json=id_file_json,
                hidden_id_file_xml=id_file_xml
            )
            form.fields['formats'].initial = 'id_formats_0'
            context = {'form': form,
                       'query_name': query_name}
            assets = {
                "js": [
                    {
                        "path": 'core_custom_queries_app/user/js/custom_queries.js',
                        "is_raw": False
                    },
                ],
            }

            return render(request,
                          "core_custom_queries_app/user/query_results.html",
                          assets=assets,
                          context=context)
        else:
            messages.add_message(request, messages.ERROR,
                                 "An internal problem occurred during the output file creation. "
                                 "The administrator has been noticed.")
            return redirect(reverse("core_custom_queries_app_index"))
    except Exception, e:
        messages.add_message(request, messages.ERROR,
                             e.message)
        return redirect(reverse("core_custom_queries_app_index"))


@decorators.permission_required(content_type=rights.custom_queries_content_type,
                                permission=rights.custom_queries_access, login_url=reverse_lazy("core_main_app_login"))
def download(request):
    """ Generate the output files to download it

    Args:
        request:

    Returns:

    """
    id_file_csv = request.POST.get('hidden_id_file_csv', None)
    id_file_json = request.POST.get('hidden_id_file_json', None)
    id_file_xml = request.POST.get('hidden_id_file_xml', None)
    name = request.POST.get('hidden_query_name', None)
    formats = request.POST.get('formats', None)

    try:
        if formats == 'csv':
            output_file = temp_output_file_api.load_file(id_file_csv)
            extension = ".csv"
            content_type = "application/CSV"
        elif formats == 'json':
            output_file = temp_output_file_api.load_file(id_file_json)
            extension = ".json"
            content_type = "application/json"
        else:
            output_file = temp_output_file_api.load_file(id_file_xml)
            extension = ".xml"
            content_type = "application/xml"
    except:
        messages.add_message(request,
                             messages.ERROR,
                             "An internal problem occurred, the administrator has been notified.")
        return redirect(reverse("custom_queries_queryChoose"))

    today = datetime.now()
    name += "_" + str(today.month) + \
            "_" + str(today.day) + \
            "_" + str(today.year) + \
            "_" + str(today.hour) + \
            "_" + str(today.minute) + \
            "_" + str(today.second)

    # FIXME: test with large files
    response = get_file_http_response(output_file, name, content_type, extension)

    return response



""" TempUserStep model
"""
from collections import defaultdict

from bson import ObjectId
from django_mongoengine import Document, fields

from core_custom_queries_app.components.dyn_query_step.models import DynQueryStep
from core_custom_queries_app.components.temp_bucket_id_files.models import TempBucketIdFiles
from core_custom_queries_app.components.temp_choice_list_file.models import TempChoiceListFile
from core_custom_queries_app.exceptions import EmptyChoicesFromQuery
from core_custom_queries_app.utils import flat_list, explore_star, is_datetime_inf_or_equal


class TempUserStep(Document):
    """
    Model which define the step associated to a user choice
    """
    step = fields.ReferenceField(DynQueryStep, blank=True)  # Base step

    # Attribute for query_management
    query_able = fields.BooleanField()  # True if the query is query able, false else
    # List of choices and files
    list_choices_id_file = fields.ListField(fields.ReferenceField(TempChoiceListFile, blank=True), blank=True)
    dict_choices_id_file = dict()  # Dict of choices and files id
    files = fields.ListField(blank=True)  # List of file for date step
    choices = fields.ListField(blank=True)  # User choices for the step
    form_type = fields.StringField()  # Form step type (radio, check, date)
    absolute_position = fields.IntField()  # Position in the list of steps for the query
    viewable_position = fields.IntField()  # Position in the list of query able steps for the query
    full_xpath = fields.StringField()  # Full path to the query

    @staticmethod
    def get_all():
        """Return all TempUserStep.

        Returns:

        """
        return TempUserStep.objects().all()

    def delete_database(self):
        """
        Delete function

        Delete each choices and the associated files.
        """

        for choices_id_file in self.list_choices_id_file:
            try:
                choices_id_file_obj = TempChoiceListFile.objects.get(id=choices_id_file.id)
                choices_id_file_obj.delete_database()
            except:
                pass

        self.delete()

    def initialize_step(self, step=None):
        """
            Initialise the query step.

            It initialize a step. If the step name is given, the information are loaded from the
            step define by the admin into step.

            :param step: The step to load.
            :return: The object initialized.
        """

        # Attribute for query_management
        self.list_choices_id_file = list()
        self.dict_choices_id_file = dict()
        self.files = list()  # List of file for date step
        self.choices = list()  # User choice for the step
        self.form_type = ''
        self.absolute_position = None
        self.viewable_position = None
        self.full_xpath = None
        self.query_able = False

        if step is not None:
            self.load_from_step_admin(step=step)

    def load_from_step_admin(self, step):
        """
            Load the step from the admin defined step.

            The information are loaded from the step define by the admin into the user's step.
            It is mainly used by the __init__.

            :param step: The step to load.
        """
        self.step = step

        self.query_able = not step.output_field

        if self.step.data_type == 'data':
            if self.step.data_min_range == 1 and self.step.data_max_range == 1:
                self.form_type = 'radio'
            else:
                self.form_type = 'check'
        else:
            self.form_type = 'date'

    def load_choices_and_xml_data(self, cursor_xml_data):
        """
            Load the dictionary of choices and files from the list of input files

            :param cursor_xml_data: List of input files
        """
        dict_choices_file = defaultdict(set)
        for file_xml_data in cursor_xml_data:
            data_dict = file_xml_data.to_mongo()
            self.get_choices(
                [data_dict['dict_content']],
                [x for x in self.full_xpath.split(".") if x],
                dict_choices_file,
                data_dict['_id']
            )
        for choices, set_files in dict_choices_file.iteritems():
            self.dict_choices_id_file[choices] = list(set_files)

        if len(self.dict_choices_id_file) == 0:
            raise EmptyChoicesFromQuery(x_path=self.full_xpath)

    def get_id_files_from_user_choices(self):
        """
        Get the list of unique files from the user choices and the list of choices and files.
        :return: List of unique files
        """
        unique_id_files = set()
        for temp_choice_list_file in self.list_choices_id_file:
            if temp_choice_list_file.choice in self.choices:
                for bucket in temp_choice_list_file.list_buckets:
                    for id_file in bucket.list_files:
                        unique_id_files.add(ObjectId(id_file))
        return list(unique_id_files)

    def get_choices_as_tuple(self):
        """
            Get list of choices.

            Get a list of possible choices for the user as tuple as
            [(choice1, choice1), (choice2, choice2), ...].

            :return: List of possible choices for the user as tuple or None if there is no choices.
        """
        if len(self.dict_choices_id_file) == 0:
            return None
        else:
            choices = set()
            for keys in self.dict_choices_id_file:
                choices.add((keys, keys))
            return sorted(tuple(choices))

    def update_choices_to_db(self):
        """
            Update the user choices for the step.
        """
        self.update(choices=self.choices)

    def update_choices_id_files_to_db(self):
        """
            Update the choices and the corresponding files to the db.

            Update the choices and the corresponding files to the db. Change also the way of data
             structure, from a dictionary based structure to a list based structure more suitable
             for database restriction.
        """
        choices_id_files_to_save = self.dict_choices_id_file

        # Delete old data in the DB
        for choice in self.list_choices_id_file:
            for bucket in choice.list_buckets:
                for files in bucket.list_files:
                    files.delete()
                bucket.delete()
            choice.delete()

        # Transform and update the data
        new_list_choices_list_files = list()
        for choice, list_value in choices_id_files_to_save.items():
            new_temp_choice_list_file = TempChoiceListFile()
            new_temp_choice_list_file.choice = choice
            list_bucket_temp = list()
            chunks = [list_value[x:x + 300] for x in xrange(0, len(list_value), 300)]
            for chunk in chunks:
                new_bucket = TempBucketIdFiles()
                new_bucket.list_files = chunk
                new_bucket.save()
                list_bucket_temp.append(new_bucket)

            new_temp_choice_list_file.list_buckets = list_bucket_temp
            new_temp_choice_list_file.save()
            new_list_choices_list_files.append(new_temp_choice_list_file)

        self.update(list_choices_id_file=new_list_choices_list_files)

    def transform_choices_files_list_to_dict(self):
        """
            Transform and return the possible choices-lists from db-list to use-dict

            Transform the choices-files list structure into a DB to a usable choices-files dict
            structure. It also delete old trace of existing choices-files dict structure.
        """
        self.dict_choices_id_file = dict()
        for list_choices_id_file in self.list_choices_id_file:
            self.dict_choices_id_file[list_choices_id_file.choice] = list()
            for buckets in list_choices_id_file.list_buckets:
                self.dict_choices_id_file[list_choices_id_file.choice].extend(buckets.list_files)

    def update_files_in_db(self):
        """
            Update the list  of files into the db
        """
        self.update(files=self.files)

    def clean_step(self):
        """
            Clean a specific step into the database.
        """
        self.update(choices=list())
        self.update(files=list())

        for list_choices_id_file in self.list_choices_id_file:
            for bucket in list_choices_id_file.list_buckets:
                bucket.delete()
            list_choices_id_file.delete()
        self.update(list_choices_id_file=list())

    def filter_user_choice_date(self, element, time_from, time_to, translate_dt):
        """
        Filter a specif element against it date value
        :param element: element too filter
        :param time_from: First threshold to filter
        :param time_to: Second threshold to filter
        :param translate_dt: Python translate for the element value
        :return: Return true is the element date is after the time_from and before the time_to
        """
        timestamp = str(element["@" + self.step.value])
        if timestamp[-1] == 'Z':
            timestamp = timestamp[:-1]
        timestamp = map(int, timestamp.translate(translate_dt).split())
        if (is_datetime_inf_or_equal(time_from, timestamp) and
                is_datetime_inf_or_equal(timestamp, time_to)) is False:
            return False
        else:
            return True

    def choose_filter_choice(self, list_elements, element_title, translate_dt, translate_bounds):
        """
        Choose a specific filter and apply it to a list of elements.
        :param list_elements: List of elements to filter
        :param element_title: Title of the elements
        :param translate_dt: Translate object for date comparison
        :param translate_bounds: Translate object for date comparison
        :return: Filtered list of elements
        """
        if self.query_able:
            if self.step.data_type == 'data':
                target = self.step.target
                choices = self.choices
                if target == 'attribute':
                    value = "@" + self.step.value
                    list_elements_cleaned = [x for x in list_elements if x[value] in choices]
                    return list_elements_cleaned
                elif target == 'name':
                    if element_title not in choices:
                        list_elements_cleaned = []
                    else:
                        list_elements_cleaned = list_elements
                    return list_elements_cleaned
                else:
                    # FIXME: #text was hardcoded, but elements don't have #text if the xml tag does not have any
                    # FIXME: attributes
                    list_elements_cleaned = [x for x in list_elements if isinstance(x, dict) and x['#text'] in choices
                                             or x in choices]
                    return list_elements_cleaned
            else:
                time_from = map(int, str(self.choices[0]).
                                translate(translate_bounds).split())
                time_to = map(int, str(self.choices[1]).
                              translate(translate_bounds).split())
                my_filter_date = self.filter_user_choice_date
                list_elements_cleaned = [
                    x
                    for x in list_elements
                    if my_filter_date(element=x, time_from=time_from, time_to=time_to,
                                      translate_dt=translate_dt)
                    ]
                return list_elements_cleaned
        else:
            return list_elements

    def get_choices(self, list_elements, list_keys_xpath, dict_choices_file, id_file):
        """
        Get possibles choices from a list of elements to explore.
        :param list_elements: List of elements to explore
        :param list_keys_xpath: List of elements from the xpath spliced
        :param dict_choices_file: Dictionary of choices, linked to a list of files
        :param id_file:Current if file
        :return:
        """
        element_title = list_keys_xpath[-1]

        while list_keys_xpath:
            key_xpath = list_keys_xpath.pop(0)
            if list_elements:
                if isinstance(list_elements[0], list):
                    list_elements = flat_list(list_elements)
            if key_xpath == "*":
                set_possibilities = explore_star(list_elements)
                for key_possible in set_possibilities:
                    list_key_param_star = [key_possible]
                    if len(list_keys_xpath) != 0:
                        list_key_param_star.extend(list_keys_xpath)
                    self.get_choices(
                        list_elements=list_elements,
                        list_keys_xpath=list_key_param_star,
                        dict_choices_file=dict_choices_file,
                        id_file=id_file
                    )
                return
            else:
                list_elements = [x.get(key_xpath) for x in list_elements if x.get(key_xpath)]

        target = self.step.target
        if target == 'name':
            dict_choices_file[element_title].add(id_file)
        else:
            if list_elements:
                if isinstance(list_elements[0], list):
                    list_elements = flat_list(list_elements)
            if target == 'attribute':
                value = "@" + self.step.value
                set_elements = (str(x.get(value)) for x in list_elements)
                for element in set_elements:
                    dict_choices_file[element].add(id_file)
            else:
                # FIXME: #text was hardcoded, but elements don't have #text if the xml tag does not have any attributes
                # set_elements = (x.get("#text") for x in list_elements)
                set_elements = (x.get("#text") if isinstance(x, dict) else str(x) for x in list_elements)
                for element in set_elements:
                    dict_choices_file[element].add(id_file)
        return




""" Dyn Query Model
"""

from django_mongoengine import fields, Document
from mongoengine import errors as mongoengine_errors
from mongoengine.queryset.base import CASCADE

from core_custom_queries_app.components.dyn_query_step.models import DynQueryStep
from core_main_app.commons import exceptions


class DynQuery(Document):
    """
    Query object created by the administrator.
    """
    # Query name
    name = fields.StringField(blank=False, unique=True)
    # Limit the query to number_records records
    is_limited = fields.BooleanField(blank=False, default=False)
    # Records number to limit the query
    number_records = fields.IntField(blank=True)
    # Group query
    group = fields.StringField(blank=False)
    # Schema associated to the query
    # FIXME: replace by a reference field
    schema = fields.StringField(blank=False)
    # List of step links to the query
    steps = fields.ListField(fields.ReferenceField(DynQueryStep,
                                                   blank=False,
                                                   reverse_delete_rule=CASCADE))

    @staticmethod
    def get_by_id(dyn_query_id):
        """Return a DynQuery given its id.

        Args:
            dyn_query_id:

        Returns:

        """
        try:
            return DynQuery.objects.get(pk=str(dyn_query_id))
        except mongoengine_errors.DoesNotExist as e:
            raise exceptions.DoesNotExist(e.message)
        except Exception as ex:
            raise exceptions.ModelError(ex.message)

    @staticmethod
    def get_all():
        """Return all DynQueries.

        Returns:

        """
        return DynQuery.objects().all()



""" Api for DynQueryStep
"""
from core_custom_queries_app.components.dyn_query_step.models import DynQueryStep


def upsert(dyn_query_step):
    """ Save or Update DynQueryStep.

    Args:
        dyn_query_step:

    Returns:

    """
    return dyn_query_step.save()


def delete(dyn_query_step):
    """

    Args:
        dyn_query_step:

    Returns:

    """
    dyn_query_step.delete()


def get_by_id(dyn_query_step_id):
    """ Return a DynQueryStep given its id.

    Args:
        dyn_query_step_id:

    Returns:

    """
    return DynQueryStep.get_by_id(dyn_query_step_id)


""" History Query api
"""
from core_custom_queries_app.models import HistoryQuery


def get_by_id(history_query_id):
    """ Return a HistoryQuery given its id.

    Args:
        history_query_id:

    Returns:

    """
    return HistoryQuery.get_by_id(history_query_id)


def get_all_by_user_id(user_id):
    """ Return a list of HistoryQuery given their user_id.

    Args:
        user_id:

    Returns:

    """
    return HistoryQuery.get_all_by_user_id(user_id)


def get_all():
    """ Return a list of all HistoryQuery.

    Returns:

    """
    return HistoryQuery.get_all()



""" Dyn Query Step Model
"""
from django_mongoengine import fields, Document
from mongoengine import errors as mongoengine_errors

from core_main_app.commons import exceptions


class DynQueryStep(Document):
    """
    Step object created by the administrator.
    """
    # Step name
    name = fields.StringField(blank=False, unique=False)
    # True if the step is not query-able, False else
    output_field = fields.BooleanField(blank=False, unique=False)
    # Relative path to the object, in dot notation
    xpath = fields.StringField(blank=False, unique=False)
    # Define the part of the node concerned by the
    target = fields.StringField(blank=False, unique=False)
    # query, could be name, attribute, or value
    value = fields.StringField(blank=True, unique=False)
    # Define the name of the attribute concerned by
    data_type = fields.StringField(blank=False, unique=False)
    # Define if the there is a restriction to the step
    restriction = fields.BooleanField(blank=True, unique=False)
    # Define the minimum range for multiple choices step
    data_min_range = fields.IntField(blank=True, unique=False)
    # Define the maximum range for multiple choices step
    data_max_range = fields.IntField(blank=True, unique=False)
    # Define the minimum range to infinity
    data_infinity = fields.BooleanField(blank=True, unique=False)      #
    # Define the number of days for date query
    date_range = fields.IntField(blank=True, unique=False)

    @staticmethod
    def get_by_id(dyn_query_step_id):
        """Return a DynQueryStep given its id.

        Args:
            dyn_query_step_id:

        Returns:

        """
        try:
            return DynQueryStep.objects.get(pk=str(dyn_query_step_id))
        except mongoengine_errors.DoesNotExist as e:
            raise exceptions.DoesNotExist(e.message)
        except Exception as ex:
            raise exceptions.ModelError(ex.message)



""" Url router for the custom queries application
"""
from django.conf.urls import url

from core_custom_queries_app.views.user import views as user_views

urlpatterns = [
    url(r'^$', user_views.choose_query,
        name='core_custom_queries_app_index'),
    url(r'^query/(?P<query_id>\w+)/steps$', user_views.query_steps,
        name='core_custom_queries_app_steps'),
    url(r'^query/(?P<history_id>\w+)/steps/recover/', user_views.recover_query_steps,
        name='core_custom_queries_app_recover_steps'),
    url(r'^query/(?P<history_id>\w+)/output/', user_views.output_files,
        name='core_custom_queries_app_output'),
    url(r'^query/download/', user_views.download,
        name='core_custom_queries_app_download'),
]





""" Custom queries exceptions
"""


class EmptyChoicesFromQuery(Exception):
    def __init__(self, x_path="", schema=""):
        if x_path != "":
            self.message = "There is no choice for the user with the xpath %s" % str(x_path)
        else:
            self.message = "There is no XML data linked to the schema %s." % str(schema)



"""
Celery tasks.
"""
from datetime import datetime

from celery.schedules import crontab
from celery.task import periodic_task
from redis import Redis, ConnectionError

from core_custom_queries_app.components.history_query import api as history_query_api
from core_custom_queries_app.components.log_file import api as log_file_api
from core_custom_queries_app.components.log_file.models import LogFile
from core_custom_queries_app.components.query_to_treat import api as query_to_treat_api
from core_custom_queries_app.components.temp_bucket_id_files import api as temp_bucket_id_files_api
from core_custom_queries_app.components.temp_bucket_output_file import api as temp_bucket_output_file_api
from core_custom_queries_app.components.temp_choice_list_file import api as temp_choice_list_file_api
from core_custom_queries_app.components.temp_output_file import api as temp_output_file_api
from core_custom_queries_app.components.temp_user_query import api as temp_user_query_api
from core_custom_queries_app.components.temp_user_step import api as temp_user_step_api
from core_custom_queries_app.utils import flat_list
from core_main_app.commons.exceptions import DoesNotExist
from qdr.settings import REDIS_URL


@periodic_task(run_every=crontab(minute=0, hour=0))
def clean_mongo_db():
    """
    Clean the database.
    :return:
    """

    # Clean the database of ended temporary query if they are ended before 7 days ago.
    datetime_now = datetime.now()
    history_queries = history_query_api.get_all()
    for history_query in history_queries:
        try:
            user_query = temp_user_query_api.get_by_id(history_query.query_id)
            if (datetime_now - user_query.last_modified).total_seconds() > (60 * 60 * 24 * 7):
                try:
                    history_query.delete_database()
                except:
                    pass
        except:
            pass

    # Delete orphan HistoryQuery
    all_history_id_to_keep = temp_user_query_api.get_all().values_list('history')
    histories_to_delete = history_query_api.get_all().filter(id__nin=all_history_id_to_keep)
    for history_to_delete in histories_to_delete:
        history_to_delete.delete_database()

    # Delete orphan TempUserQuery
    all_temp_user_query_to_keep = history_query_api.get_all().values_list('query_id')
    all_temp_user_query_to_delete = temp_user_query_api.get_all().filter(id__nin=all_temp_user_query_to_keep)
    for query_to_delete in all_temp_user_query_to_delete:
        query_to_delete.delete_database()

    # Delete orphan TempOutputFile
    temp_output_files_id_to_keep = list()
    all_temp_user_queries = temp_user_query_api.get_all()
    for temp_user_query in all_temp_user_queries:
        if temp_user_query.str_id_file_json != "":
            temp_output_files_id_to_keep.append(temp_user_query.str_id_file_json)
        if temp_user_query.str_id_file_xml != "":
            temp_output_files_id_to_keep.append(temp_user_query.str_id_file_xml)
        if temp_user_query.str_id_file_csv != "":
            temp_output_files_id_to_keep.append(temp_user_query.str_id_file_csv)
    temp_output_files_to_delete = temp_output_file_api.get_all().filter(
        id__nin=temp_output_files_id_to_keep
    )
    for temp_output_file_to_delete in temp_output_files_to_delete:
        temp_output_file_to_delete.delete_database()

    # Delete orphan TempBucketOutputFiles
    temp_bucket_output_files_id_to_keep = list()
    all_files_temp_bucket_output_files_to_keep = temp_output_file_api.get_all().values_list('file')
    all_files_temp_bucket_output_files_to_keep = flat_list(all_files_temp_bucket_output_files_to_keep)
    for id_file in all_files_temp_bucket_output_files_to_keep:
        temp_bucket_output_files_id_to_keep.append(id_file.id)
    temp_bucket_output_files_to_delete = temp_bucket_output_file_api.get_all().filter(
        id__nin=temp_bucket_output_files_id_to_keep)
    for temp_bucket_output_file_to_delete in temp_bucket_output_files_to_delete:
        temp_bucket_output_file_to_delete.delete_database()

    # Delete orphan TempUserStep
    all_temp_user_step_id_to_keep = list()
    all_temp_user_step_to_keep = temp_user_query_api.get_all().values_list('list_steps')
    all_temp_user_step_to_keep = flat_list(all_temp_user_step_to_keep)
    for temp_user_step_to_keep in all_temp_user_step_to_keep:
        all_temp_user_step_id_to_keep.append(temp_user_step_to_keep.id)
    all_temp_user_steps_to_delete = temp_user_step_api.get_all().filter(id__nin=all_temp_user_step_id_to_keep)
    for step_to_delete in all_temp_user_steps_to_delete:
        step_to_delete.delete_database()

    # Delete orphan TempChoiceListFile
    all_temp_choice_list_file_id_to_keep = list()
    all_temp_choice_list_file_to_keep = temp_user_step_api.get_all().values_list('list_choices_id_file')
    all_temp_choice_list_file_to_keep = flat_list(all_temp_choice_list_file_to_keep)
    for temp_choice_list_file_to_keep in all_temp_choice_list_file_to_keep:
        all_temp_choice_list_file_id_to_keep.append(temp_choice_list_file_to_keep.id)
    all_temp_choice_list_file_to_delete = \
        temp_choice_list_file_api.get_all().filter(id__nin=all_temp_choice_list_file_id_to_keep)
    for temp_choice_list_file_to_delete in all_temp_choice_list_file_to_delete:
        temp_choice_list_file_to_delete.delete_database()

    # Delete orphan TempBucketIdFiles
    all_temp_bucket_id_files = list()
    all_temp_bucket_id_files_to_keep = temp_bucket_id_files_api.get_all().values_list('list_files')
    all_temp_bucket_id_files_to_keep = flat_list(all_temp_bucket_id_files_to_keep)
    for temp_bucket_id_files_to_keep in all_temp_bucket_id_files_to_keep:
        all_temp_bucket_id_files.append(temp_bucket_id_files_to_keep)
    all_temp_bucket_id_files_to_delete = temp_bucket_id_files_api.get_all().filter(id__nin=all_temp_bucket_id_files)
    for temp_bucket_id_files_to_delete in all_temp_bucket_id_files_to_delete:
        temp_bucket_id_files_to_delete.delete_database()


@periodic_task(run_every=crontab(minute='*'))
def get_files_to_create():
    """
    Treat the queries in the waiting queue.

    Get the queries from the database and put them into the waiting list.
    If a the waiting list is empty or the worker is already run, do nothing, else, launch the worker.
    """
    redis_server = Redis.from_url(REDIS_URL)

    queries_to_treat = query_to_treat_api.get_all()

    try:
        for query_to_treat in queries_to_treat:
            Redis.rpush(redis_server, 'list_ids', str(query_to_treat.query.id))
            query_to_treat.delete_database()

        if not Redis.exists(redis_server, 'is_working'):
            Redis.set(redis_server, 'is_working', False)

        if Redis.get(redis_server, 'is_working') == 'False':
            Redis.set(redis_server, 'is_working', True)
            if Redis.exists(redis_server, 'list_ids'):
                run_worker(redis_server)
            Redis.set(redis_server, 'is_working', False)
    except ConnectionError, e:
        log_file = LogFile(application="Custom Queries",
                           message="Redis not reachable, is it running?",
                           additionalInformation={'message': e.message},
                           timestamp=datetime.now())
        log_file_api.upsert(log_file)


def run_worker(redis_server):
    """
    Worker algorithm.

    While the waiting list is not empty, treat the query in order.

    :param redis_server: Redis server instance.
    """
    try:
        while Redis.exists(redis_server, 'list_ids'):
            current_id = Redis.lpop(redis_server, 'list_ids')
            redis_server.set('current_id', current_id)
            try:
                query = temp_user_query_api.get_by_id(current_id)
                query.create_outputs_file()
            except DoesNotExist:
                pass
            redis_server.set('current_id', "None")
    except ConnectionError, e:
        log_file = LogFile(application="Custom Queries",
                           message="Redis not reachable, is it running?",
                           additionalInformation={'message': e.message},
                           timestamp=datetime.now())
        log_file_api.upsert(log_file)



""" Apps file for setting core package when app is ready
"""
from django.apps import AppConfig


class CustomQueriesAppConfig(AppConfig):
    """ Core application settings.
    """
    name = 'core_custom_queries_app'

    def ready(self):
        """ Run when the app is ready.

        Returns:

        """
        import core_custom_queries_app.permissions.discover as discover
        discover.init_permissions()



""" TempOutputFile api
"""
from core_custom_queries_app.components.temp_output_file.models import TempOutputFile


def load_file(id_output_file):
    """ Load output file from its id

    Args:
        id_output_file:

    Returns:

    """
    return TempOutputFile.load_file(id_output_file)


def get_all():
    """Return all TempOutputFile.

    Returns:

    """
    return TempOutputFile.get_all()



"""
Describe the admin forms used by custom_queries
"""

import re

from django import forms

from core_custom_queries_app.components.dyn_query import api as dyn_query_api
from core_main_app.components.template import api as template_api
from core_main_app.components.template_version_manager import api as template_version_manager_api


class QueryPieceForm(forms.Form):
    """
    Form to create or modify query steps
    """

    # Possible type of data
    typesData = (
        ('data', 'Data'),
        ('date', 'Date-Time')
    )

    # Possible type of target
    typesTarget = (
        ('attribute', 'Attribute'),
        ('name', 'Name'),
        ('value', 'Value')
    )

    # Step id - Hidden field
    step_id = forms.CharField(widget=forms.HiddenInput(),
                              label="Id",
                              required=False)

    # Step name - Text field
    name = forms.CharField(label="Name",
                           required=True,
                           widget=forms.TextInput(attrs={'class': 'full_item'})
                           )

    # Step xpath - Text field
    xpath = forms.CharField(label="XPath",
                            required=True,
                            widget=forms.TextInput(attrs={'class': 'full_item'})
                            )

    # Step output field, False if query able - Checkbox field
    output_field = forms.BooleanField(label="Output field only",
                                      required=False,
                                      widget=forms.CheckboxInput(attrs={'class': 'full_item'})
                                      )

    # Step target, represent the part of the node which will be query - Select field
    target = forms.ChoiceField(label="Target",
                               required=False,
                               choices=typesTarget,
                               initial='attribute',
                               widget=forms.Select(attrs={'class': 'full_item'})
                               )

    # Attribute name which will be query if target is "attribute" - Text field
    value = forms.CharField(label="Attribute name",
                            required=False,
                            widget=forms.TextInput(attrs={'class': 'full_item'})
                            )

    # Data type query, data or date - Select field
    data_type = forms.ChoiceField(label="Data type",
                                  required=False,
                                  choices=typesData,
                                  widget=forms.Select(attrs={'class': 'full_item'})
                                  )

    # Restriction, if there is applicable restriction to the query - Checkbox field
    restriction = forms.BooleanField(label="Restriction",
                                     required=False,
                                     initial=False,
                                     widget=forms.CheckboxInput(attrs={'class': 'full_item'})
                                     )

    # Minimum data range if applicable - Text field
    data_min_range = forms.IntegerField(label="Minimum range",
                                        required=False,
                                        widget=forms.TextInput(attrs={'class': 'full_item', 'disabled': 'disabled'})
                                        )

    # Maximum data range if applicable - Integer field
    data_max_range = forms.IntegerField(label="Maximum range",
                                        required=False,
                                        widget=forms.TextInput(attrs={'class': 'full_item', 'disabled': 'disabled'})
                                        )

    # Infinite data range if applicable - Checkbox field
    data_infinity = forms.BooleanField(label="Infinity",
                                       required=False,
                                       widget=forms.CheckboxInput(attrs={'class': 'full_item', 'disabled': 'disabled'})
                                       )

    # Date range if applicable - Integer field
    date_range = forms.IntegerField(label="Maximum day range",
                                    required=False,
                                    widget=forms.TextInput(attrs={'class': 'full_item', 'disabled': 'disabled'})
                                    )

    def __init__(self, *args, **kwargs):
        super(QueryPieceForm, self).__init__(*args, **kwargs)
        self.initial['data_type'] = 'data'
        self.empty_permitted = False

    def clean(self):
        # Define the cleaning rules for the step
        if 'name' not in self.cleaned_data or self.cleaned_data['name'] == "":
            if 'name' in self.errors:
                del self.errors['name']
            self.add_error('name', "You have to fill the name.")
        else:
            name = self.cleaned_data['name']
            if not (re.match('^[a-zA-Z0-9_]+$', name) is not None):
                self.add_error('name',
                               "A query name could only be composed by alphanumeric "
                               "characters.")

        if 'xpath' not in self.cleaned_data or self.cleaned_data['xpath'] == "":
            if 'xpath' in self.errors:
                del self.errors['xpath']
            self.add_error('xpath', "You have to fill the xpath.")
        else:
            xpath = self.cleaned_data["xpath"]
            if ".." in xpath:
                self.add_error('xpath', "The xpath can not contain \"..\".")

        if self.cleaned_data['target'] == 'attribute' \
                and self.cleaned_data['output_field'] is False:
            if 'value' not in self.cleaned_data or self.cleaned_data['value'] == '':
                self.add_error('value', "You have to fill the name of the attribute you want to "
                                        "query.")
        if self.cleaned_data['restriction'] is True \
                and self.cleaned_data['data_type'] == 'data' \
                and self.cleaned_data['output_field'] is False:
            if 'data_min_range' not in self.cleaned_data:
                if 'data_min_range' in self.errors:
                    del self.errors['data_min_range']
                self.add_error('data_min_range', "You have to fill the minimum range and has to be"
                                                 " an integer.")
            elif self.cleaned_data['data_min_range'] == '' or self.cleaned_data['data_min_range']\
                    is None:
                self.add_error('data_min_range', "You have to fill the minimum range.")
            elif self.cleaned_data['data_min_range'] < 0:
                self.add_error('data_min_range', "The minimum range has to be a positive integer.")

        if self.cleaned_data['data_infinity'] is False and self.cleaned_data['restriction'] is True \
                and self.cleaned_data['data_type'] == 'data' \
                and self.cleaned_data['output_field'] is False:
            if 'data_max_range' not in self.cleaned_data:
                if 'data_max_range' in self.errors:
                    del self.errors['data_max_range']
                self.add_error('data_max_range', "You have to fill the maximum range and has to be "
                                                 "an integer.")
            elif self.cleaned_data['data_max_range'] == '' or self.cleaned_data['data_max_range']\
                    is None:

                self.add_error('data_max_range', "You have to fill the maximum range.")
            elif self.cleaned_data['data_max_range'] < 0:
                self.add_error('data_max_range', "The maximum range has to be a positive integer.")

        if self.cleaned_data['data_type'] == 'data' \
                and self.cleaned_data['output_field'] is False:
            if self.cleaned_data['restriction'] is True:
                if self.cleaned_data['data_infinity'] is False:
                    if 'data_min_range' in self.cleaned_data and 'data_max_range' in self.cleaned_data:
                        if type(self.cleaned_data['data_max_range']) == int and \
                                        type(self.cleaned_data['data_min_range']) == int:
                            if self.cleaned_data['data_max_range'] < self.cleaned_data['data_min_range']:
                                self.add_error('data_max_range',
                                               "The maximum range has to superior to the minimum range.")

        if self.cleaned_data['restriction'] is True \
                and self.cleaned_data['data_type'] == 'date' \
                and self.cleaned_data['output_field'] is False:
            if 'date_range' not in self.cleaned_data:
                if 'date_range' in self.errors:
                    del self.errors['date_range']
                self.add_error('date_range', "You have to fill the date range and has to be an integer.")
            elif self.cleaned_data['date_range'] == '' or self.cleaned_data['date_range'] is None:
                self.add_error('date_range', "You have to fill the date range.")
            elif self.cleaned_data['date_range'] <= 0:
                self.add_error('date_range', "The date has to be a positive integer.")


class Query(forms.Form):
    """
        Form to create or modify query
    """

    # Possible group for the query
    groups = (
        ('admin', 'Admin-only'),
        ('user', 'User-allowed'),
    )

    # Query ID
    id = forms.CharField(
        widget=forms.HiddenInput(),
        label="Id",
        required=False
    )

    # Query name
    name = forms.CharField(label="Query name",
                           required=True)

    # Is query limited to number_records records
    is_limited = forms.BooleanField(label="Is the query limited",
                                    required=False,
                                    initial=False)

    # Number of records
    number_records = forms.IntegerField(
        label="Number of records to limit the query",
        required=False,
        min_value=1
    )

    # Schema ID associated to the query
    schema = forms.ChoiceField(label="Schema",
                               required=True)

    # Group associated to the query
    group = forms.ChoiceField(label="Group",
                              required=True,
                              choices=groups,
                              widget=forms.RadioSelect(),
                              initial='admin')

    # Delete field
    to_delete = forms.CharField(widget=forms.HiddenInput(attrs={'id': "to_delete"}),
                                initial=1234,
                                required=False)

    def _post_clean(self):
        """
        Cleaning method for POST call.
        """
        if 'schema' in self.errors:
            del self.errors['schema'][0]
            self.add_error('schema', 'The previously selected schema is not available anymore, '
                                     'please select an other.')

    def __init__(self, *args, **kwargs):
        """
        Init function for query.
        :param args: Used by the super class init.
        :param kwargs: Used by the super class init.
        """
        super(Query, self).__init__(*args, **kwargs)
        self.initial['group'] = 'admin'

        template_version_managers = template_version_manager_api.get_active_global_version_manager()

        list_templates = list()

        for template_version_manager in template_version_managers:
            for version_id in template_version_manager.versions:
                template = template_api.get(version_id)
                list_templates.append([version_id, template.display_name])

        self.fields['schema'].choices = list_templates

    def clean(self):
        """
        Clean function for the query
        :return:
        """
        if 'name' not in self.cleaned_data:
            if 'name' in self.errors:
                del self.errors['name']
            self.add_error('name', "The query name is required.")
        else:
            name = self.cleaned_data['name']
            if not (re.match('^[a-zA-Z0-9_]+$', name) is not None):
                self.add_error('name',
                               "A query name could only be composed by alphanumeric "
                               "characters.")

            cur_id = self.cleaned_data['id']
            if cur_id != '':
                names = dyn_query_api.get_all().filter(id__nin=[self.cleaned_data["id"]]).values_list('name')
            else:
                names = dyn_query_api.get_all().values_list('name')
            if name == "":
                self.add_error('name', "The name can not be empty.")
            elif name in names:
                self.add_error('name', "A query with this name has already been registered.")

        if 'is_limited' in self.cleaned_data:
            if self.cleaned_data['is_limited']:
                if 'number_records' in self.cleaned_data:
                    number_records = self.cleaned_data["number_records"]
                    if isinstance(number_records, int):
                        if number_records <= 0:
                            self.add_error('number_records', "The number of records has to be a positive integer ")
                    else:
                        self.add_error('number_records', "The number of records has to be a positive integer.")
                else:
                    self.add_error('number_records', "The number of records has to be filled")


class LogFilesFormPreview(forms.Form):
    """
    Form preview log files, give basic information
    """

    # Log file id - Hidden field
    id = forms.CharField(required=True,
                         widget=forms.widgets.HiddenInput())

    # Application link to the log file - Text field
    application = forms.CharField(label='Application',
                                  required=True,
                                  widget=forms.TextInput(attrs={'class': 'full_item', 'readonly': 'true'})
                                  )

    # Timestamp link to the log file - Text field
    timestamp = forms.DateTimeField(required=True,
                                    widget=forms.TextInput(attrs={'class': 'full_item', 'readonly': 'true'})
                                    )

    # Log file message - Text field
    message = forms.CharField(label='Error message',
                              required=True,
                              widget=forms.TextInput(attrs={'class': 'full_item', 'readonly': 'true'})
                              )


class LogFilesFormDetails(forms.Form):
    """
        Form detail log files, give full information
    """

    # Log file id - Hidden field
    id = forms.CharField(required=True,
                         widget=forms.widgets.HiddenInput())

    # Application link to the log file - Text field
    application = forms.CharField(label='Application',
                                  required=True,
                                  widget=forms.TextInput(attrs={'class': 'full_item', 'readonly': 'true'})
                                  )

    # Timestamp link to the log file - Text field
    timestamp = forms.DateTimeField(required=True,
                                    widget=forms.TextInput(attrs={'class': 'full_item', 'readonly': 'true'})
                                    )

    # Log file code - Text field
    code = forms.CharField(label='Error code',
                           required=False,
                           widget=forms.TextInput(attrs={'class': 'full_item', 'readonly': 'true'})
                           )

    # Log file message - Text field
    message = forms.CharField(label='Error message',
                              required=True,
                              widget=forms.TextInput(attrs={'class': 'full_item', 'readonly': 'true'})
                              )

    # Log file additional information, presented by a dictionary - Textarea field
    additional = forms.CharField(label='Additional information',
                                 required=False,
                                 widget=forms.Textarea(attrs={'class': 'additional_element', 'readonly': 'true'})
                                 )



"""
Describe all the view admin-only
"""
from django.contrib import messages
from django.contrib.admin.views.decorators import staff_member_required
from django.forms import formset_factory
from django.shortcuts import redirect
from django.urls import reverse

from core_custom_queries_app.components.dyn_query import api as dyn_query_api
from core_custom_queries_app.components.dyn_query.models import DynQuery
from core_custom_queries_app.components.dyn_query_step import api as dyn_query_step_api
from core_custom_queries_app.components.dyn_query_step.models import DynQueryStep
from core_custom_queries_app.components.log_file import api as log_file_api
from core_custom_queries_app.views.admin.forms import Query, QueryPieceForm, LogFilesFormPreview, LogFilesFormDetails
from core_main_app.commons.exceptions import DoesNotExist, ModelError
from core_main_app.components.template import api as template_api
from core_main_app.components.template_version_manager import api as template_version_manager_api
from core_main_app.utils.pagination.django_paginator.results_paginator import ResultsPaginator
from core_main_app.utils.rendering import admin_render


@staff_member_required
def queries_list(request):
    """
    View list of queries defined by the administrator.
    Used by the admin.
    Create a form where all the queries are listed.

    Args:
        request:

    Returns:

    """
    queries = dyn_query_api.get_all()

    query_list = []
    for query in queries:
        query_dict = dict()
        query_dict['id'] = query.pk
        query_dict['name'] = query.name
        try:
            template = template_api.get(query.schema)
            template_version_manager = template_version_manager_api.get_by_version_id(str(template.id))
            query_dict['schema'] = template_version_manager.title
            if template_version_manager.is_disabled:
                query_dict['schema'] += '-deleted'
        except Exception, e:
            query_dict['schema'] = 'Template deleted from the database.'

        query_dict['group'] = query.group

        query_list.append(query_dict)

    queries = query_list

    assets = {
        "css": ['core_custom_queries_app/admin/css/queries_list.css'],
    }

    return admin_render(request,
                        "core_custom_queries_app/admin/queries_list.html",
                        assets=assets,
                        context={"queries": queries})


@staff_member_required
def query_builder(request):
    """
    Create a query. Only used by an administrators.
    A query is composed by a common part (DynQuery) and at least one step (DynQueryStep).
    If the query has been validated  the query is saved into the database.
    A new query can not have the same name as another into the database.
    If not, the query is sent to the user for modification.

    Args:
        request:

    Returns:

    """
    if request.method == 'POST':
        # Take the data from the forms
        form = Query(request.POST, auto_id=True)
        number_forms = int(request.POST['form-TOTAL_FORMS'])
        query_piece_form = formset_factory(QueryPieceForm, extra=number_forms)
        formset = query_piece_form(request.POST, request.FILES)

        # Validate the forms
        if form.is_valid():
            if formset.is_valid():
                try:
                    # Create the query
                    data = form.cleaned_data
                    query = DynQuery(name=data["name"], group=data["group"], schema=data["schema"],
                                     is_limited=data["is_limited"], number_records=data["number_records"])

                    # Create the query steps
                    list_query_steps = []
                    for fs in formset:
                        data = fs.cleaned_data
                        step = DynQueryStep(name=data["name"], xpath=data["xpath"], output_field=data["output_field"],
                                            target= data["target"], value=data["value"], data_type=data["data_type"],
                                            restriction=data["restriction"], data_min_range=data["data_min_range"],
                                            data_max_range=data["data_max_range"], data_infinity=data["data_infinity"],
                                            date_range=data["date_range"])
                        list_query_steps.append(step)
                        try:
                            dyn_query_step_api.upsert(step)
                        except Exception, e:
                            for step in query.steps:
                                dyn_query_step_api.delete(step)
                            messages.add_message(request, messages.ERROR, e.message)
                            break

                    # Add the steps to the query
                    query["steps"] = list_query_steps
                    try:
                        dyn_query_api.upsert(query)
                    except Exception, e:
                        for step in query.steps:
                            dyn_query_step_api.delete(step)
                        messages.add_message(request, messages.ERROR, e.message)

                    messages.add_message(request, messages.SUCCESS, 'Query saved.')
                    return redirect(reverse("admin:core_custom_queries_app_queries"))

                except Exception as e:
                    messages.add_message(request, messages.ERROR, e.message)
    else:
        # Create an empty form to create a query
        form = Query()
        query_piece_form = formset_factory(QueryPieceForm)
        formset = query_piece_form()

    context = {'form': form,
               'formset': formset,
               'query_piece_form': query_piece_form}

    assets = {
        "js": [
            {
                "path": 'core_custom_queries_app/admin/js/custom_queries.js',
                "is_raw": False
            },
        ],
    }

    return admin_render(request,
                        'core_custom_queries_app/admin/query_builder.html',
                        assets=assets,
                        context=context)


# FIXME: refactor with query_builder
@staff_member_required
def edit_query(request, query_id):
    """
    Update or delete a query from the query selection.
    Used by the admin.
    Create a form with the query's information. The modification are saved if the form is validated.
    A confirmation pop-up is shown is the query is about to be deleted.
    Come back to the query selection if the query has been deleted.

    Args:
        request:
        query_id:

    Returns:

    """
    # Come from the form
    if request.method == 'POST':
        # Take the data from the forms
        # FIXME: using the same form as the one for a new query prevent from updating query without changing its name
        form = Query(request.POST, auto_id=True)
        number_forms = int(request.POST['form-TOTAL_FORMS'])
        query_piece_form = formset_factory(QueryPieceForm, extra=number_forms)
        formset = query_piece_form(request.POST, request.FILES)

        # Validate the forms
        if form.is_valid():
            if formset.is_valid():
                form_data = form.cleaned_data
                formset_data = formset.cleaned_data
                try:
                    query = dyn_query_api.get_by_id(query_id)
                except DoesNotExist:
                    messages.add_message(request, messages.INFO,
                                         'The query you are trying to manage does not exist anymore.')
                    return redirect(reverse("admin:core_custom_queries_app_queries"))
                except ModelError, e:
                    messages.add_message(request, messages.WARNING, e.message)
                    return redirect(reverse("admin:core_custom_queries_app_queries"))

                query.name = form_data["name"]
                query.is_limited = form_data["is_limited"]
                query.group = form_data["group"]
                query.schema = form_data["schema"]
                query.number_records = form_data["number_records"]
                dyn_query_api.upsert(query)

                # Create the query steps
                for step_data in formset_data:
                    try:
                        step = dyn_query_step_api.get_by_id(step_data["step_id"])
                    except DoesNotExist:
                        messages.add_message(request, messages.INFO,
                                             'The query you are trying to manage does not exist anymore.')
                    except ModelError, e:
                        messages.add_message(request, messages.WARNING, e.message)

                    step.name=step_data.get("name")
                    step.xpath=step_data.get("xpath")
                    step.output_field=step_data.get("output_field")
                    step.target=step_data.get("target")
                    step.value=step_data.get("value")
                    step.data_type=step_data.get("data_type")
                    step.restriction=step_data.get("restriction")
                    step.data_min_range=step_data.get("data_min_range")
                    step.data_max_range=step_data.get("data_max_range")
                    step.data_infinity=step_data.get("data_infinity")
                    step.date_range=step_data.get("date_range")
                    dyn_query_step_api.upsert(step)

                messages.add_message(request, messages.SUCCESS, 'Query updated.')
                return redirect(reverse("admin:core_custom_queries_app_queries"))

    # Come from the query selector
    else:
        try:
            query = dyn_query_api.get_by_id(query_id)
        except DoesNotExist:
            messages.add_message(request, messages.INFO, "The matching query does not exist.")
            return redirect(reverse("admin:core_custom_queries_app_queries"))
        except ModelError, e:
            messages.add_message(request, messages.WARNING, e.message)
            return redirect(reverse("admin:core_custom_queries_app_queries"))

        nb_steps = len(query.steps)

        # Load the query
        param_query = {
            'id': query.id,
            'name': query.name,
            'is_limited': query.is_limited,
            'number_records': query.number_records,
            'group': query.group,
            'schema': query.schema
        }

        form = Query(param_query)

        query_piece_form = formset_factory(QueryPieceForm, extra=nb_steps)

        # Load the query steps
        data = dict()
        data['form-TOTAL_FORMS'] = nb_steps
        data['form-INITIAL_FORMS'] = nb_steps
        data['form-MAX_NUM_FORMS'] = nb_steps
        position = 0
        for step in query.steps:
            data['form-' + str(position) + '-step_id'] = step.id
            data['form-' + str(position) + '-name'] = step["name"]
            data['form-' + str(position) + '-xpath'] = step["xpath"]
            data['form-' + str(position) + '-output_field'] = step["output_field"]
            data['form-' + str(position) + '-target'] = step["target"]
            data['form-' + str(position) + '-value'] = step["value"]
            data['form-' + str(position) + '-data_type'] = step["data_type"]
            data['form-' + str(position) + '-restriction'] = step["restriction"]
            data['form-' + str(position) + '-data_min_range'] = step["data_min_range"]
            data['form-' + str(position) + '-data_max_range'] = step["data_max_range"]
            data['form-' + str(position) + '-data_infinity'] = step["data_infinity"]
            data['form-' + str(position) + '-date_range'] = step["date_range"]
            position += 1

        formset = query_piece_form(data)

    context = {'form': form,
               'formset': formset,
               'query_piece_form': query_piece_form,
               'query_id': query_id}

    assets = {
        "js": [
            {
                "path": 'core_custom_queries_app/admin/js/custom_queries.js',
                "is_raw": False
            },
        ],
    }

    modals = ['core_custom_queries_app/admin/custom_queries/modals/delete_query.html']

    return admin_render(request,
                        'core_custom_queries_app/admin/update_query.html',
                        modals=modals,
                        assets=assets,
                        context=context)


@staff_member_required
def list_errors(request):
    """View and delete log files from the database.
    Used by the admin.
    Create a form set with the log files information. The log file could be deleted.
    The log files are viewed with pagination.
    If no log files are in the database, an error message will be shown.

    :param request: Django HttpRequest
        _no Params at first instance, page from pagination if page change
        _POST dictionary when come back from admin_list_errors web page with the log files which
         will be deleted.
    """

    # Query(ies) witch will be deleted
    if request.method == "POST":
        number_forms = int(request.POST['form-TOTAL_FORMS'])
        log_file_formset = formset_factory(LogFilesFormPreview, extra=number_forms, can_delete=True)
        formset = log_file_formset(request.POST, request.FILES)

        if formset.is_valid():
            for form in formset.deleted_forms:
                data = form.cleaned_data
                if 'DELETE' in data:
                    id_object = data['id']
                    try:
                        logfile = log_file_api.get_by_id(id_object)
                        logfile.delete_database()
                    except:
                        pass

    # Get the log files
    log_file_list = log_file_api.get_all().order_by("-timestamp")

    # Get the page if page change
    page = request.GET.get('page', 1)
    result_paginator = ResultsPaginator()
    log_files = result_paginator.get_results(log_file_list, page)
    number_log_files = len(log_files)

    # Create a form set with the log files' information
    log_file_formset = formset_factory(LogFilesFormPreview, extra=number_log_files, can_delete=True)

    log_file_details_formset = formset_factory(LogFilesFormDetails, extra=number_log_files)

    data = dict()
    data['form-TOTAL_FORMS'] = number_log_files
    data['form-INITIAL_FORMS'] = number_log_files
    data['form-MAX_NUM_FORMS'] = number_log_files

    data_details = dict()
    data_details['form-TOTAL_FORMS'] = number_log_files
    data_details['form-INITIAL_FORMS'] = number_log_files
    data_details['form-MAX_NUM_FORMS'] = number_log_files

    position = 0

    for log_file in log_files:
        additional = ''
        for key, value in log_file['additionalInformation'].iteritems():
            additional += str(key) + ": " + str(value) + "\r\n"

        data_details['form-' + str(position) + '-id'] = str(log_file["id"])
        data_details['form-' + str(position) + '-timestamp'] = log_file["timestamp"]
        data_details['form-' + str(position) + '-code'] = log_file["code"]
        data_details['form-' + str(position) + '-application'] = log_file["application"]
        data_details['form-' + str(position) + '-message'] = log_file["message"]
        data_details['form-' + str(position) + '-additional'] = additional

        data['form-' + str(position) + '-id'] = str(log_file["id"])
        data['form-' + str(position) + '-timestamp'] = log_file["timestamp"]
        data['form-' + str(position) + '-application'] = log_file["application"]
        data['form-' + str(position) + '-message'] = log_file["message"]
        data['form-' + str(position) + '-DELETE'] = ''

        position += 1

    formset = log_file_formset(data)
    formset_detail = log_file_details_formset(data_details)

    if len(log_files) == 0:
        messages.add_message(request, messages.INFO, "There is no error to show.")

    assets = {
        "js": [
            {
                "path": 'core_custom_queries_app/admin/js/custom_queries.js',
                "is_raw": False
            },
        ],
    }
    modals = ['core_custom_queries_app/admin/log_files/modals/edit_modals.html']

    context = {
        'number_log_files': number_log_files,
        'log_file_formset': log_file_formset,
        'formset': formset,
        'log_files': log_files,
        'log_file_details_formset': log_file_details_formset,
        'formset_detail': formset_detail
    }

    return admin_render(request,
                        'core_custom_queries_app/admin/listing_errors.html',
                        assets=assets,
                        modals=modals,
                        context=context)



""" Add Custom Queries to main menu
"""
from django.core.urlresolvers import reverse
from menu import Menu, MenuItem


Menu.add_item(
    "main", MenuItem("Query Repository", reverse("core_custom_queries_app_index"))
)

types_children = (
    MenuItem("Custom Queries List", reverse("admin:core_custom_queries_app_queries"), icon="list"),
    MenuItem("View Error Logs", reverse("admin:core_custom_queries_app_errors"), icon="list"),
)

Menu.add_item(
    "admin", MenuItem("CUSTOM QUERIES", None, children=types_children)
)



""" TempChoiceListFile model
"""
from django_mongoengine import Document, fields

from core_custom_queries_app.components.temp_bucket_id_files.models import TempBucketIdFiles


class TempChoiceListFile(Document):
    """
    Model which associate a specific choice the list of files
    """
    choice = fields.StringField()
    list_buckets = fields.ListField(fields.ReferenceField(TempBucketIdFiles, blank=True), blank=True)

    @staticmethod
    def get_all():
        """Return all TempChoiceListFile.

        Returns:

        """
        return TempChoiceListFile.objects().all()

    def delete_database(self):
        """
        Delete function

        Delete each part of the choice and the associated files.
        """
        for bucket in self.list_buckets:
            try:
                bucket_obj = TempBucketIdFiles.objects.get(id=bucket.id)
                bucket_obj.delete_database()
            except:
                pass

        self.delete()



""" Forms for Custom Queries user views
"""

from django import forms


class FormChooseQuery(forms.Form):
    """
    Form to choose a query with radio selection.
    """

    # Query field - Radio field
    query = forms.ChoiceField(widget=forms.widgets.RadioSelect(), required=True)

    def __init__(self, *args, **kwargs):
        _query = kwargs.pop('query', None)
        super(FormChooseQuery, self).__init__(*args, **kwargs)
        self.fields['query'].choices = _query


class FormBase(forms.Form):
    """
    Base information for steps
    """

    # Query name - Hidden field.
    queryName = forms.CharField(label='Query Name', required=True, widget=forms.widgets.HiddenInput())
    # Query position - Hidden field.
    position = forms.IntegerField(label='Position in query', required=True, widget=forms.widgets.HiddenInput())
    # Query id - Hidden field.
    user_query = forms.CharField(label='User query', required=True, widget=forms.widgets.HiddenInput())
    # Form type - Hidden field.
    data_type = forms.CharField(label='Query step type', required=True, widget=forms.widgets.HiddenInput())
    # Hidden input if the user want to go back.
    step_back = forms.BooleanField(required=False, widget=forms.widgets.HiddenInput())
    # History id linked to the step - Hidden field.
    history_id = forms.CharField(required=True, widget=forms.widgets.HiddenInput())

    def __init__(self, *args, **kwargs):
        _queryName = kwargs.pop('queryName', None)
        _position = kwargs.pop('position', None)
        _user_query = kwargs.pop('user_query', None)
        _data_type = kwargs.pop('data_type', None)
        _step_back = kwargs.pop('step_back', None)
        _history_id = kwargs.pop('history_id', None)
        super(FormBase, self).__init__(*args, **kwargs)
        self.fields['queryName'].initial = _queryName
        self.fields['position'].initial = _position
        self.fields['user_query'].initial = _user_query
        self.fields['data_type'].initial = _data_type
        self.fields['step_back'].initial = _step_back
        self.fields['history_id'].initial = _history_id


class FormRadio(FormBase):
    """
    Radio form for a query step.
    """

    # Choices for the current step - Radio field
    choices = forms.ChoiceField(widget=forms.widgets.RadioSelect(), required=True)

    def __init__(self, *args, **kwargs):
        # Init radio form
        _choices = kwargs.pop('choices', None)
        _stepName = kwargs.pop('stepName', None)
        super(FormRadio, self).__init__(*args, **kwargs)
        self.fields['choices'].choices = _choices
        self.fields['choices'].label = _stepName
        self.fields['data_type'].data = 'radio'


class FormCheck(FormBase):
    """
    Check form for the query step
    """

    # Choices for the current step - Check field
    choices = forms.MultipleChoiceField(widget=forms.widgets.CheckboxSelectMultiple(), required=True)
    # Minimum range for the current step - Hidden field
    data_min_range = forms.IntegerField(widget=forms.widgets.HiddenInput())
    # Maximum range for the current step - Hidden field
    data_max_range = forms.IntegerField(widget=forms.widgets.HiddenInput())
    # Infinity for the current step - Hidden field
    data_infinity = forms.BooleanField(widget=forms.widgets.HiddenInput())

    def __init__(self, *args, **kwargs):
        # Init check forms
        _choices = kwargs.pop('choices', None)
        _dataMinRange = kwargs.pop('data_min_range', None)
        _dataMaxRange = kwargs.pop('data_max_range', None)
        _dataInfinity = kwargs.pop('data_infinity', None)
        _stepName = kwargs.pop('stepName', None)
        super(FormCheck, self).__init__(*args, **kwargs)
        self.fields['choices'].choices = _choices
        self.fields['choices'].label = _stepName
        self.fields['data_min_range'].initial = _dataMinRange
        self.fields['data_max_range'].initial = _dataMaxRange
        self.fields['data_infinity'].initial = _dataInfinity


class FormDateTime(FormBase):
    """
    Date time form
    """

    # First threshold, "From" date
    timeFrom = forms.DateTimeField(label='From',
                                   required=True,
                                   widget=forms.TextInput(
                                       attrs={'readonly': 'True', 'class': 'form-control'}
                                   ))

    # Second threshold, "To" date
    timeTo = forms.DateTimeField(label='To',
                                 required=True,
                                 widget=forms.TextInput(
                                     attrs={'readonly': 'True', 'class': 'form-control'}
                                 ))

    # Number of days query able, maximum difference between "to" and "from"
    date_range = forms.CharField(widget=forms.widgets.HiddenInput())

    def __init__(self, *args, **kwargs):
        _dateRange = kwargs.pop('date_range', None)
        super(FormDateTime, self).__init__(*args, **kwargs)
        self.fields['date_range'].initial = _dateRange
        self.fields['data_type'].initial = 'date'


class FormResult(forms.Form):
    """
    Result form, with file preview and file download.
    """

    # Different format files
    choices = (
        ('csv', 'ASCII CSV'),
        ('json', 'JSON'),
        ('xml', 'XML')
    )

    # CSV file id - Hidden field
    hidden_id_file_csv = forms.CharField(required=False, widget=forms.widgets.HiddenInput())
    # JSON file id - Hidden field
    hidden_id_file_json = forms.CharField(required=False, widget=forms.widgets.HiddenInput())
    # XML file id - Hidden field
    hidden_id_file_xml = forms.CharField(required=False, widget=forms.widgets.HiddenInput())
    # CSV file preview - Hidden field
    preview_file_csv = forms.CharField(required=False, widget=forms.widgets.HiddenInput())
    # JSON file preview - Hidden field
    preview_file_json = forms.CharField(required=False, widget=forms.widgets.HiddenInput())
    # XML file preview - Hidden field
    preview_file_xml = forms.CharField(required=False, widget=forms.widgets.HiddenInput())
    # Query name - Hidden field
    hidden_query_name = forms.CharField(required=True, widget=forms.widgets.HiddenInput())
    # Format file selected - Radio field
    formats = forms.ChoiceField(required=False,
                                choices=choices,
                                widget=forms.widgets.RadioSelect(),
                                initial='csv')

    def __init__(self, *args, **kwargs):
        # Init result forms
        hidden_id_file_csv = kwargs.pop('hidden_id_file_csv', None)
        hidden_id_file_json = kwargs.pop('hidden_id_file_json', None)
        hidden_id_file_xml = kwargs.pop('hidden_id_file_xml', None)
        preview_file_csv = kwargs.pop('preview_file_csv', None)
        preview_file_json = kwargs.pop('preview_file_json', None)
        preview_file_xml = kwargs.pop('preview_file_xml', None)
        hidden_query_name = kwargs.pop('queryName', None)
        super(FormResult, self).__init__(args, kwargs)
        self.data['hidden_id_file_csv'] = hidden_id_file_csv
        self.data['hidden_id_file_json'] = hidden_id_file_json
        self.data['hidden_id_file_xml'] = hidden_id_file_xml
        self.data['preview_file_csv'] = preview_file_csv
        self.data['preview_file_json'] = preview_file_json
        self.data['preview_file_xml'] = preview_file_xml
        self.data['hidden_query_name'] = hidden_query_name
        self.fields['formats'].initial = 'csv'


class FormHistory(forms.Form):
    """
    Form to select / delete history query
    """

    # Query ID - Hidden field
    id_history = forms.CharField(widget=forms.widgets.HiddenInput())
    # Query name - Text field
    query_name = forms.CharField(
        label='Query name',
        required=False,
        widget=forms.TextInput(attrs={'class': 'full_item'})
    )
    # Last modified timestamp - Text field
    query_last_modified = forms.DateTimeField(
        label='Last modified',
        required=False,
        widget=forms.TextInput(attrs={'class': 'full_item'})
    )
    # Query message - Text field
    query_message = forms.CharField(
        label='Status',
        required=False,
        widget=forms.TextInput(attrs={'class': 'full_item'})
    )
    # Query status - Hidden fields
    status = forms.IntegerField(widget=forms.widgets.HiddenInput())



""" TempBucketIdFiles model
"""
from django_mongoengine import Document, fields


class TempBucketIdFiles(Document):
    """
    Part of the input file
    """
    list_files = fields.ListField(blank=True)

    @staticmethod
    def get_all():
        """Return all TempBucketIdFiles.

        Returns:

        """
        return TempBucketIdFiles.objects().all()

    def delete_database(self):
        """
        Delete function
        """
        self.delete()



""" TempBucketIdFiles
"""
from core_custom_queries_app.components.temp_bucket_id_files.models import TempBucketIdFiles


def get_all():
    """Return all TempBucketIdFiles.

    Returns:

    """
    return TempBucketIdFiles.get_all()



""" LogFile Model
"""
from django_mongoengine import fields, Document
from mongoengine import errors as mongoengine_errors

from core_main_app.commons import exceptions


class LogFile(Document):
    """
    Error reports for administrators
    """
    application = fields.StringField(blank=False)            # Django application name concerned by the error
    code = fields.IntField(blank=True)                       # Error code
    message = fields.StringField(blank=False)                # Error message
    additionalInformation = fields.DictField(blank=True)     # Dictionary of additional information
    timestamp = fields.DateTimeField(blank=True)             # Error creation date

    @staticmethod
    def get_by_id(log_file_id):
        """Return a LogFile given its id.

        Args:
            log_file_id:

        Returns:

        """
        try:
            return LogFile.objects.get(pk=str(log_file_id))
        except mongoengine_errors.DoesNotExist as e:
            raise exceptions.DoesNotExist(e.message)
        except Exception as ex:
            raise exceptions.ModelError(ex.message)

    @staticmethod
    def get_all():
        """Return all LogFiles.

        Returns:

        """
        return LogFile.objects().all()

    def delete_database(self):
        """
        Delete function
        """
        self.delete()



""" AJAX views for admin custom queries
"""
from django.contrib import messages
from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpResponseBadRequest, HttpResponse
from django.urls import reverse

from core_custom_queries_app.components.dyn_query import api as dyn_query_api
from core_main_app.commons.exceptions import DoesNotExist


@staff_member_required
def delete_query(request):
    try:
        query_id = request.POST['id']
        try:
            query = dyn_query_api.get_by_id(query_id)
            dyn_query_api.delete(query)
        except DoesNotExist:
            messages.add_message(request, messages.INFO,
                                 'The query you are trying to manage does '
                                 'not exist anymore.')
            return HttpResponseBadRequest()
        messages.add_message(request, messages.SUCCESS, 'Query deleted.')
        return HttpResponse(reverse("admin:core_custom_queries_app_queries"))

    except Exception, e:
        return HttpResponseBadRequest(e.message, content_type='application/javascript')



""" TempUserStep api.
"""
from core_custom_queries_app.components.temp_user_step.models import TempUserStep


def get_all():
    """Return all TempUserStep.

    Returns:

    """
    return TempUserStep.get_all()



""" TempUserQuery api
"""
from core_custom_queries_app.models import TempUserQuery


def get_by_id(temp_user_query_id):
    """Return a TempUserQuery given its id.

    Args:
        temp_user_query_id:

    Returns:

    """
    return TempUserQuery.get_by_id(temp_user_query_id)


def get_by_history_id(history_id):
    """Return a TempUserQuery given its history id.

    Args:
        history_id:

    Returns:

    """
    return TempUserQuery.get_by_history_id(history_id)


def get_all():
    """Return all TempUserQuery.

    Returns:

    """
    return TempUserQuery.get_all()



"""
Python functions used in the algorithm.
"""
from collections import defaultdict
from itertools import chain, ifilterfalse
from json import dumps

from core_main_app.utils.notifications.mail import send_mail


def get_dict_element_header(title=None, k=None, v=None):
    """
    File creation function - Get data for output file
    :param title: Element title
    :param k: Key value
    :param v: Value associated to the key
    :return: Dictionary with the [title], the key and the value
    """
    if title:
        return {
            "title": title,
            "key": k,
            "value": v
        }
    else:
        return {
            "key": k,
            "value": v
        }


def get_header_parents(dict_key):
    """
    Get information from parent elements.
    :param dict_key: Dictionary of keys.
    :return: List of dictionaries composed by parent element.
    """
    title = dict_key.get("#title")
    if len(dict_key) == 1:
        return [
            get_dict_element_header(title)
        ]
    else:
        return [
            get_dict_element_header(title, k, v)
            for k, v in dict_key.iteritems()
            if k != "#title"
            ]


def get_title_data_leaf(list_leaves, map_keys):
    """
    Get title for leaf elements.
    :param list_leaves: List of leaf elements.
    :param map_keys: Map of hash and elements.
    """
    ddict_title_data = defaultdict(list)
    for hash_leaf in list_leaves:
        leaf = map_keys[hash_leaf]
        title = leaf.pop("#title")
        ddict_title_data[title].append(leaf)
    return ddict_title_data


def get_common_key_and_specific_header(list_leaves, map_keys):
    """
    Get all the common key and value in the dict_leaves.
    :param list_leaves: List of leaves
    :param map_keys: Map of hash and elements.
    :return: List of unique keys and list of specific headers
    """
    dict_key_set_value = dict()
    list_keys = list()
    first = True

    for key_hash in list_leaves:
        dict_leaf = map_keys[key_hash]
        if first:
            for k_first, v_first in dict_leaf.iteritems():
                dict_key_set_value[k_first] = v_first
            list_keys = dict_key_set_value.keys()
            first = False
        else:
            for k, v in dict_leaf.iteritems():
                if k in list_keys \
                        and k in dict_key_set_value \
                        and v != dict_key_set_value[k]:
                    dict_key_set_value.pop(k)
    node_name = None
    if "#title" in dict_key_set_value.keys():
        node_name = dict_key_set_value["#title"]
    list_keys_headers_common = [
        (k, get_dict_element_header(node_name, k, v))
        for k, v in dict_key_set_value.iteritems()]
    list_keys_unique = list()
    list_specific_header = list()
    for keys_headers_common in list_keys_headers_common:
        list_keys_unique.append(keys_headers_common[0])
        list_specific_header.append(keys_headers_common[1])
    return list_keys_unique, list_specific_header


def print_bloc(list_specific_header, ddict_title_data, common_keys,
               list_file_xml, list_file_json, list_file_csv):
    """
    File creation function - Print data into output file

    :param list_specific_header: List of specific headers.
    :param ddict_title_data: Default dictionary of title
    :param common_keys: Common keys between the list of specific elements
    :param list_file_xml: Different parts of xml file.
    :param list_file_json: Different parts of json file.
    :param list_file_csv: Different parts of csv file.
    """

    print_headers(list_specific_header, list_file_xml, list_file_json,
                  list_file_csv)

    for title_leaf, list_dict_leaf in ddict_title_data.iteritems():
        print_leaves(title_leaf, list_dict_leaf, common_keys, list_file_xml, list_file_json,
                     list_file_csv)

    list_file_json[-1] = list_file_json[-1][:-3] + list_file_json[-1][-2:]  # Delete last ,
    list_file_json.append(3 * "\t" + "},\r\n")
    list_file_xml.append("\t</Item>\r\n")


def print_headers(list_headers, list_file_xml, list_file_json, list_file_csv):
    """
    File creation function - Print header into output file

    :param list_headers: List of headers
    :param list_file_xml: Different parts of xml file.
    :param list_file_json: Different parts of json file.
    :param list_file_csv: Different parts of csv file.
    """
    list_particular_header_csv = list()
    list_particular_header_json = list()
    list_particular_header_xml = ["\t<Item "]

    list_file_json.append(3 * "\t" + "{\r\n")

    for header in list_headers:
        if header["key"] is not None:
            if "title" in header:
                if header["key"] != "#title":
                    list_particular_header_csv.append("_".join((header["title"], header["key"], str(header["value"]))))
                    list_particular_header_json.append(": ".join(
                        (
                            "\"-" + str(header["title"])
                            + "_"
                            + str(header["key"]).replace("@", "", 1)
                            + "\"",
                            "\"" + str(header["value"]) + "\""
                        )
                    ))
                    list_particular_header_xml.append("=".join(
                        (
                            str(header["title"])
                            + "_"
                            + str(header["key"]).replace("@", "", 1),
                            "\"" + str(header["value"]) + "\""
                        )
                    ))

            elif header["key"] != "#title":
                list_particular_header_csv.append("_".join((str(header["key"]), str(header["value"]))))
                list_particular_header_json.append(": ".join(
                    (
                        "\"-" + str(header["key"]).replace("@", "", 1) + "\"",
                        "\"" + str(header["value"]) + "\""
                    )
                ))

                list_particular_header_xml.append("=".join(
                    (
                        str(header["key"]).replace("@", "", 1),
                        "\"" + str(header["value"]) + "\""
                    )
                ))
        else:
            list_particular_header_csv.append(str(header["title"]))

    list_file_csv.append(", ".join(list_particular_header_csv) + "\r\n")
    list_file_json.append("\t\t\t\t" + ",\r\n\t\t\t\t".join(list_particular_header_json) + ",\r\n")
    list_file_xml.append(" ".join(list_particular_header_xml) + ">\r\n")


def print_leaves(title_leaf, list_dict_leaf, common_keys, list_file_xml, list_file_json, list_file_csv):
    """
    File creation function - Print leaf element into output file

    :param title_leaf: Title of the leaf
    :param list_dict_leaf: Element of the leaf
    :param common_keys: List of common keys
    :param list_file_xml: Different parts of xml file.
    :param list_file_json: Different parts of json file.
    :param list_file_csv: Different parts of csv file.
    :return:
    """
    common_keys_to_avoid = []
    if len(common_keys) != 1:
        common_keys_to_avoid = common_keys

    list_file_json.append(4 * "\t" + "\"" + title_leaf + "\": [\r\n")
    for leaf in list_dict_leaf:
        list_file_csv.append("," + ", ".join(
            [
                title_leaf + ": ".join((str(k_leaf), str(v_leaf)))
                for k_leaf, v_leaf in leaf.iteritems()
                if k_leaf not in common_keys_to_avoid
                ]
        ) + "\r\n")

        list_file_json.append(5 * "\t" + "{\r\n" + 6 * "\t" + ",\r\n\t\t\t\t\t\t".join(
            [
                ": ".join(
                    (
                        "\"" + str(k_leaf).replace("@", "-") + "\"",
                        dumps(str(v_leaf))
                    )
                )
                for k_leaf, v_leaf in leaf.iteritems()
                if k_leaf not in common_keys_to_avoid
                ]
        ) + "\r\n" + 5 * "\t" + "},\r\n" )

        list_file_xml.append(2 * "\t" + "<" + title_leaf + " " + " ".join(
            [
                "=".join(
                    (
                        str(k_leaf).replace("@", ""),
                        "\"" + str(v_leaf) + "\""
                    )
                )
                for k_leaf, v_leaf in leaf.iteritems()
                if k_leaf not in common_keys_to_avoid and k_leaf != "#text"
            ]))

        if "#text" in leaf:
            list_file_xml.append("> ")
            list_file_xml.append(str(leaf["#text"]))
            list_file_xml.append(" </")
            list_file_xml.append(title_leaf)
            list_file_xml.append(">\r\n")
        else:
            list_file_xml.append("/>\r\n")

    list_file_json[-1] = list_file_json[-1][:-3] + list_file_json[-1][-2:]  # Delete last ","
    list_file_json.append(4 * "\t" + "],\r\n")


# Gathered data function
def flat_list(list_elements):
    """
    Flat a list of list into a list: [[1,2,3],[4,5,6]] ==> [1,2,3,4,5,6]

    :param list_elements: List of elements
    :return: List of flattened lists.
    """
    return list(chain.from_iterable(list_elements))


def explore_star(list_elements):
    """
    Get list of possible element to solve "*".
    :param list_elements: List of element to solve.
    :return: List of elements to replace the star.
    """
    return ifilterfalse(lambda x: x[0] in ['_', '@', "#"], set(chain.from_iterable(list_elements)))


def get_general_key_output_dictionary(element_cleaned, element_title):
    """
    Create a dictionary from a node attribute. The title is added.
    :param element_cleaned: Element cleaned to transform.
    :param element_title: Element title.
    :return: Element dictionary.
    """
    dict_key_general = {x: y for x, y in element_cleaned.iteritems() if x[0] in ("@", "#")}
    dict_key_general["#title"] = element_title
    return dict_key_general


def get_list_keys_from_xpath(xpath):
    """
    Split an xpath by the ".".
    :param xpath: the XPath to split.
    :return: List of XPath parts.
    """
    list_keys = xpath.split(".")
    if list_keys[-1] == "":
        list_keys = list_keys[:-1]
    if list_keys[0] == "":
        list_keys = list_keys[1:]

    return list_keys


def is_datetime_inf_or_equal(date_ref, date_comp):
    """
        Compare if date_ref <= date_comp

        Compare efficiently if date_ref <= date_comp. The date_comp is duplicated to allow the
        chaining function: is datetime_inf_or_equal(a, b), datetime_inf_or_equal(b, c) ...

        :param date_ref: Reference date, format: [Year, Month, Day, Hour, Minute, Second]
        :param date_comp: Compare date, format: [Year, Month, Day, Hour, Minute, Second]
        :return: True if date_ref <= date_comp, False if date_ref > date_comp
    """
    i = 0
    while date_ref[i] == date_comp[i] and i != 4:
        i += 1

    if date_ref[i] < date_comp[i] or i == 4:
        return True
    return False


# Prepare data for DB query
def possible_projection(str_projection):
    """
    Get the possible projection.
    "a.b.c"         ==> "a.b.c"
    "a.b.c.*"       ==> "a.b.c"
    "a.b.c.*.*"     ==> "a.b.c"
    "a.b.c.@title"  ==> "a.b.c"
    :param str_projection: String
    :return:
    """
    str_possible_projection = (str_projection.split("*")[0].split("@title")[0])
    if str_possible_projection.endswith("."):
        return str_possible_projection[:-1]
    else:
        return str_possible_projection


def send_mail_query_end(user, query):
    """
    Send the mail end query to the user.

    :param user: User to send the mail to
    :param query: Ended query
    """
    context = {
        'lastname': user.last_name,
        'firstname': user.first_name,
        'query_name': query.query.name
    }

    send_mail(subject='Query ended',
              path_to_template='core_custom_queries_app/user/email/end_query.html',
              context=context, recipient_list=[user.email])



"""
Url file used by django for admin's views.
"""
from django.contrib import admin
from django.conf.urls import url

from core_custom_queries_app.views.admin import views as admin_views
from core_custom_queries_app.views.admin import ajax as admin_ajax

admin_urls = [
    url(r'^custom_queries/list$', admin_views.queries_list,
        name='core_custom_queries_app_queries'),
    url(r'^custom_queries/query_builder$', admin_views.query_builder,
        name='core_custom_queries_app_query_builder'),
    url(r'^custom_queries/(?P<query_id>\w+)/edit$', admin_views.edit_query,
        name='core_custom_queries_app_edit_query'),
    url(r'^custom_queries/list_errors$', admin_views.list_errors,
        name='core_custom_queries_app_errors'),

    url(r'^custom_queries/delete$', admin_ajax.delete_query,
        name='core_custom_queries_app_delete_query'),
]

urls = admin.site.get_urls()
admin.site.get_urls = lambda: admin_urls + urls



""" TempBucketOutputFiles model
"""
from django_mongoengine import Document, fields


class TempBucketOutputFiles(Document):
    """
    Part of the output file string
    """
    piece_file = fields.StringField()

    @staticmethod
    def get_all():
        """Return all TempBucketOutputFiles.

        Returns:

        """
        return TempBucketOutputFiles.objects().all()

    def delete_database(self):
        """
        Delete function
        """
        self.delete()



""" Api for DynQueries
"""
from core_custom_queries_app.components.dyn_query.models import DynQuery
from core_custom_queries_app.components.dyn_query_step import api as dyn_query_step_api
from core_custom_queries_app.components.temp_user_query import api as temp_user_query_api


def upsert(dyn_query):
    """ Save or Update DynQuery.

    Args:
        dyn_query:

    Returns:

    """
    return dyn_query.save()


def get_by_id(dyn_query_id):
    """ Return a DynQuery given its id.

    Args:
        dyn_query_id:

    Returns:

    """
    return DynQuery.get_by_id(dyn_query_id)


def get_all():
    """ Get all DynQueries.

    Returns:

    """
    return DynQuery.get_all()


def delete(dyn_query):
    """ Delete DynQuery.

    Args:
        dyn_query:

    Returns:

    """
    # delete linked steps
    for step in dyn_query.steps:
        dyn_query_step_api.delete(step)

    # delete linked temp user queries
    temp_queries = temp_user_query_api.get_all().filter(query=dyn_query)
    for temp_query in temp_queries:
        temp_query.delete_database()
    dyn_query.delete()



""" TempBucketOutputFile api
"""
from core_custom_queries_app.components.temp_bucket_output_file.models import TempBucketOutputFiles


def get_all():
    """Return all TempBucketOutputFiles.

    Returns:

    """
    return TempBucketOutputFiles.get_all()
